class RobotFrameworkFacade(object):
    """
    Nothing on Robot Framework is currently typed, so, this is a facade
    to help to deal with it so that we don't add lots of things to ignore its
    imports/typing.
    """

    @property
    def get_model(self):
        from robot.api import get_model  # type:ignore

        return get_model

    @property
    def TestSuite(self):
        from robot.api import TestSuite

        return TestSuite

    @property
    def Token(self):
        from robot.api import Token

        return Token

    @property
    def DataError(self):
        from robot.errors import DataError  # type:ignore

        return DataError

    @property
    def EXECUTION_CONTEXTS(self):
        from robot.running.context import EXECUTION_CONTEXTS  # type:ignore

        return EXECUTION_CONTEXTS

    @property
    def SettingsBuilder(self):
        from robot.running.builder.transformers import SettingsBuilder  # type:ignore

        return SettingsBuilder

    @property
    def SuiteBuilder(self):
        from robot.running.builder.transformers import SuiteBuilder  # type:ignore

        return SuiteBuilder

    @property
    def TestDefaults(self):
        from robot.running.builder.testsettings import TestDefaults  # type:ignore

        return TestDefaults

    def get_libraries_imported_in_namespace(self):
        EXECUTION_CONTEXTS = self.EXECUTION_CONTEXTS
        return set(EXECUTION_CONTEXTS.current.namespace._kw_store.libraries)

    def run_test_body(self, context, test):
        from robot import version

        IS_ROBOT_4_ONWARDS = not version.get_version().startswith("3.")
        if IS_ROBOT_4_ONWARDS:
            from robot.running.bodyrunner import BodyRunner  # noqa

            BodyRunner(context, templated=False).run(test.body)
        else:
            from robot.running.steprunner import StepRunner  # noqa

            StepRunner(context, False).run_steps(test.keywords.normal)

    @property
    def EmbeddedArgumentsHandler(self):
        from robot.running.userkeyword import EmbeddedArgumentsHandler  # type:ignore

        return EmbeddedArgumentsHandler



from robocorp_ls_core.basic import implements
from robocorp_ls_core.constants import NULL
from robocorp_ls_core.protocols import IDirCache, check_implements, T
from robocorp_ls_core.robotframework_log import get_logger

from collections import namedtuple
from pathlib import Path
from typing import Any, Generic, Callable, Optional
import functools
import os


log = get_logger(__name__)


def instance_cache(func):
    """
    Use as decorator:

    class MyClass(object):

        @instance_cache
        def cache_this(self):
            ...

    Clear the cache with:

    instance = MyClass()
    instance.cache_this()
    instance.cache_this.cache_clear(instance)

    # set_lock(instance, lock) can be used to make the cache thread-safe.
    # i.e.:
    instance.cache_this.set_lock(instance, lock)

    """
    cache_key = "__cache__%s" % (func.__name__,)
    cache_key_lock = "__cache__%s_lock" % (func.__name__,)

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):
        from robocorp_ls_core.protocols import Sentinel

        try:
            cache = getattr(self, "__instance_cache__")
        except:
            cache = {}
            setattr(self, "__instance_cache__", cache)

        lock = cache.get(cache_key_lock, NULL)
        with lock:
            try:
                func_cache = cache[cache_key]
            except KeyError:
                func_cache = cache[cache_key] = {}

            args_cache_key = None
            if args:
                args_cache_key = (args_cache_key, tuple(args))
            if kwargs:
                # We don't do that because if the caller uses args and then
                # later kwargs, we'd have to match the parameter to the position,
                # so, simplify for now and don't accept kwargs.
                raise AssertionError("Cannot currently deal with kwargs.")

            ret = func_cache.get(args_cache_key, Sentinel)
            if ret is Sentinel:
                ret = func(self, *args, **kwargs)
                func_cache[args_cache_key] = ret
            return ret

    def set_lock(self, lock):
        # Note that this will make sure we'll pre-create the lock so that
        # the __instance_cache__ is always there (so, there should be
        # no race condition for it).
        try:
            cache = getattr(self, "__instance_cache__")
        except:
            cache = {}
            setattr(self, "__instance_cache__", cache)

        cache[cache_key_lock] = lock

    def cache_clear(self):
        try:
            cache = getattr(self, "__instance_cache__")
        except:
            pass
        else:
            lock = cache.get(cache_key_lock, NULL)
            with lock:
                cache.pop(cache_key, None)

    new_func.cache_clear = cache_clear
    new_func.set_lock = set_lock
    return new_func


class DirCache(object):
    """
    To be used as:

    dir_cache = DirCache(cache_directory)
    dir_cache.store("some_key", 1)
    value = dir_cache.load("some_key", int) # Ok

    try:
        value = dir_cache.load("some_key", dict)
    except KeyError:
        ... error: value is not a dict

    try:
        dir_cache.load("key does not exist", dict)
    except KeyError:
        ... error, key does not exist
    """

    def __init__(self, cache_directory: str) -> None:
        self._cache_dir = cache_directory
        os.makedirs(self._cache_dir, exist_ok=True)

    def _encode_key(self, key: Any) -> str:
        import hashlib

        return hashlib.sha224(repr(key).encode("utf-8")).hexdigest()

    def _get_file_for_key(self, key: Any) -> str:
        key_encoded = self._encode_key(key)
        return os.path.join(self._cache_dir, key_encoded)

    @implements(IDirCache.store)
    def store(self, key, value):
        import json

        with open(self._get_file_for_key(key), "w") as stream:
            stream.write(json.dumps({"key": key, "value": value}))

    @implements(IDirCache.load)
    def load(self, key, expected_class):
        import json

        filename = self._get_file_for_key(key)
        if not os.path.exists(filename):
            raise KeyError(f"Key: {key} not found in cache.")

        try:
            with open(filename, "r") as stream:
                contents = json.loads(stream.read())
            value = contents["value"]

        except Exception:
            msg = f"Unable to load key: {key} from cache."
            log.debug(msg)
            raise KeyError(msg)

        if not isinstance(value, expected_class):
            raise KeyError(
                f"Unable to load key: {key} from cache (expected it to be a {expected_class} was {type(value)}."
            )
        return value

    @implements(IDirCache.discard)
    def discard(self, key):
        filename = self._get_file_for_key(key)
        try:
            os.remove(filename)
        except Exception:
            pass

    def __typecheckself__(self) -> None:
        _: IDirCache = check_implements(self)


CachedFileMTimeInfo = namedtuple("CachedFileMTimeInfo", "st_mtime, st_size, path")


class CachedFileInfo(Generic[T]):
    def __init__(self, file_path: Path, compute_value: Callable[[Path], T]):
        self._file_path = file_path
        self._mtime_info: Optional[CachedFileMTimeInfo] = self._get_mtime_cache_info(
            file_path
        )
        # Note that we only get the value after getting the mtime (so, the
        # constructor receives a callable and not the value so that the cache
        # has no risk of having stale values).
        self._value = compute_value(file_path)

    @property
    def value(self) -> T:
        return self._value

    @property
    def file_path(self) -> Path:
        return self._file_path

    def _get_mtime_cache_info(self, file_path: Path) -> Optional[CachedFileMTimeInfo]:
        """
        Cache based on the time/size of a given path.
        """
        try:
            stat = file_path.stat()
            return CachedFileMTimeInfo(stat.st_mtime, stat.st_size, str(file_path))
        except:
            # Probably removed in the meanwhile.
            log.exception(f"Unable to get mtime for: {file_path}")
            return None

    def is_cache_valid(self) -> bool:
        return self._mtime_info == self._get_mtime_cache_info(self.file_path)



from itertools import chain

from robot.api.parsing import ModelTransformer, If, IfHeader, ElseHeader, ElseIfHeader, Token, KeywordCall, Comment, End

try:
    from robot.api.parsing import ReturnStatement, Break, Continue, InlineIfHeader
except ImportError:
    ReturnStatement, Break, Continue, InlineIfHeader = None, None, None, None

from robotidy.utils import ROBOT_VERSION, normalize_name, get_comments, flatten_multiline
from robotidy.decorators import check_start_end_line


class InlineIf(ModelTransformer):
    """
    Replaces IF blocks with inline IF.

    It will only replace IF block if it can fit in one line shorter than `line_length` (default 80) parameter and return
    variables matches for all ELSE and ELSE IF branches.

    Following code::

        *** Test Cases ***
        Test
            IF    $condition1
                Keyword    argument
            END
            IF    $condition2
                ${var}  Keyword
            ELSE
                ${var}  Keyword 2
            END
            IF    $condition1
                Keyword    argument
                Keyword 2
            END

    will be transformed to:

        *** Test Cases ***
        Test
            IF    $condition1    Keyword    argument
            ${var}    IF    $condition2    Keyword    ELSE    Keyword 2
            IF    $condition1
                Keyword    argument
                Keyword 2
            END

    Too long inline IFs (over `line_length` character limit) will be replaced with normal IF block.
    You can decide to not replace IF blocks containing ELSE or ELSE IF branches by setting `skip_else` to True.

    Supports global formatting params: `--startline` and `--endline`.

    See https://robotidy.readthedocs.io/en/latest/transformers/InlineIf.html for more examples.
    """

    ENABLED = ROBOT_VERSION.major >= 5

    def __init__(self, line_length: int = 80, skip_else: bool = False):
        self.line_length = line_length
        self.skip_else = skip_else

    @check_start_end_line
    def visit_If(self, node: If):  # noqa
        if node.errors or getattr(node.end, "errors", None):
            return node
        if self.is_inline(node):
            return self.handle_inline(node)
        self.generic_visit(node)
        if node.orelse:
            self.generic_visit(node.orelse)
        if self.no_end(node):
            return node
        indent = node.header.tokens[0]
        if not (self.should_transform(node) and self.assignment_identical(node)):
            return node
        return self.to_inline(node, indent.value)

    def should_transform(self, node):
        if node.header.errors:
            return False
        if (
            len(node.body) > 1
            or not node.body
            or not isinstance(node.body[0], (KeywordCall, ReturnStatement, Break, Continue))
        ):
            return False
        if node.orelse:
            return self.should_transform(node.orelse)
        return True

    @staticmethod
    def if_to_branches(if_block):
        while if_block:
            yield if_block
            if_block = if_block.orelse

    def assignment_identical(self, node):
        else_found = False
        assigned = []
        for branch in self.if_to_branches(node):
            if isinstance(branch.header, ElseHeader):
                else_found = True
            if not isinstance(branch.body[0], KeywordCall) or not branch.body[0].assign:
                assigned.append([])
            else:
                assigned.append([normalize_name(assign).replace("=", "") for assign in branch.body[0].assign])
            if len(assigned) > 1 and assigned[-1] != assigned[-2]:
                return False
        if any(x for x in assigned):
            return else_found
        return True

    def is_shorter_than_limit(self, inline_if):
        line_len = sum(self.if_len(branch) for branch in self.if_to_branches(inline_if))
        return line_len <= self.line_length

    @staticmethod
    def no_end(node):
        if not node.end:
            return True
        if not len(node.end.tokens) == 1:
            return False
        return not node.end.tokens[0].value

    @staticmethod
    def is_inline(node):
        return isinstance(node.header, InlineIfHeader)

    @staticmethod
    def if_len(if_st):
        return sum(
            len(tok.value)
            for tok in chain(if_st.body[0].tokens if if_st.body else [], if_st.header.tokens)
            if tok.value != "\n"
        )

    def to_inline(self, node, indent):
        tail = node
        comments = self.collect_comments_from_if(indent, node)
        if_block = head = self.inline_if_from_branch(node, indent)
        while tail.orelse:
            if self.skip_else:
                return node
            tail = tail.orelse
            comments += self.collect_comments_from_if(indent, tail)
            head.orelse = self.inline_if_from_branch(tail, self.formatting_config.separator)
            head = head.orelse
        if self.is_shorter_than_limit(if_block):
            return (*comments, if_block)
        return node

    def inline_if_from_branch(self, node, indent):
        if not node:
            return None
        separator = self.formatting_config.separator
        last_token = Token(Token.EOL) if node.orelse is None else Token(Token.SEPARATOR, separator)
        assigned = None

        if isinstance(node.body[0], KeywordCall):
            assigned = node.body[0].assign
            keyword = self.to_inline_keyword(node.body[0], separator, last_token)
        elif isinstance(node.body[0], ReturnStatement):
            keyword = self.to_inline_return(node.body[0], separator, last_token)
        elif isinstance(node.body[0], Break):
            keyword = Break(self.to_inline_break_continue_tokens(Token.BREAK, separator, last_token))
        elif isinstance(node.body[0], Continue):
            keyword = Continue(self.to_inline_break_continue_tokens(Token.CONTINUE, separator, last_token))
        else:
            return node

        # check for ElseIfHeader first since it's child of IfHeader class
        if isinstance(node.header, ElseIfHeader):
            header = ElseIfHeader(
                [Token(Token.ELSE_IF), Token(Token.SEPARATOR, separator), Token(Token.ARGUMENT, node.header.condition)]
            )
        elif isinstance(node.header, IfHeader):
            tokens = [Token(Token.SEPARATOR, indent)]
            if assigned:
                for assign in assigned:
                    tokens.extend([Token(Token.ASSIGN, assign), Token(Token.SEPARATOR, separator)])
            tokens.extend(
                [
                    Token(Token.INLINE_IF),
                    Token(Token.SEPARATOR, separator),
                    Token(Token.ARGUMENT, node.header.condition),
                ]
            )
            header = InlineIfHeader(tokens)
        elif isinstance(node.header, ElseHeader):
            header = ElseHeader([Token(Token.ELSE)])
        else:
            return node
        return If(header=header, body=[keyword])

    @staticmethod
    def to_inline_keyword(keyword, separator, last_token):
        tokens = [Token(Token.SEPARATOR, separator), Token(Token.KEYWORD, keyword.keyword)]
        for arg in keyword.get_tokens(Token.ARGUMENT):
            tokens.extend([Token(Token.SEPARATOR, separator), arg])
        tokens.append(last_token)
        return KeywordCall(tokens)

    @staticmethod
    def to_inline_return(node, separator, last_token):
        tokens = [Token(Token.SEPARATOR, separator), Token(Token.RETURN_STATEMENT)]
        for value in node.values:
            tokens.extend([Token(Token.SEPARATOR, separator), Token(Token.ARGUMENT, value)])
        tokens.append(last_token)
        return ReturnStatement(tokens)

    @staticmethod
    def to_inline_break_continue_tokens(token, separator, last_token):
        return [Token(Token.SEPARATOR, separator), Token(token), last_token]

    @staticmethod
    def join_on_separator(tokens, separator):
        for token in tokens:
            yield token
            yield separator

    @staticmethod
    def collect_comments_from_if(indent, node):
        comments = get_comments(node.header.tokens)
        for statement in node.body:
            comments += get_comments(statement.tokens)
        if node.end:
            comments += get_comments(node.end)
        return [Comment.from_params(comment=comment.value, indent=indent) for comment in comments]

    def create_keyword_for_inline(self, kw_tokens, indent, assign):
        keyword_tokens = []
        for token in kw_tokens:
            keyword_tokens.append(Token(Token.SEPARATOR, self.formatting_config.separator))
            keyword_tokens.append(token)
        return KeywordCall.from_tokens(
            [
                Token(Token.SEPARATOR, indent + self.formatting_config.separator),
                *assign,
                *keyword_tokens[1:],
                Token(Token.EOL),
            ]
        )

    def flatten_if_block(self, node):
        node.header.tokens = flatten_multiline(
            node.header.tokens, self.formatting_config.separator, remove_comments=True
        )
        for index, statement in enumerate(node.body):
            node.body[index].tokens = flatten_multiline(
                statement.tokens, self.formatting_config.separator, remove_comments=True
            )
        return node

    def is_if_multiline(self, node):
        for branch in self.if_to_branches(node):
            if branch.header.get_token(Token.CONTINUATION):
                return True
            if any(statement.get_token(Token.CONTINUATION) for statement in branch.body):
                return True
        return False

    def flatten_inline_if(self, node):
        indent = node.header.tokens[0].value
        comments = self.collect_comments_from_if(indent, node)
        node = self.flatten_if_block(node)
        head = node
        while head.orelse:
            head = head.orelse
            comments += self.collect_comments_from_if(indent, head)
            head = self.flatten_if_block(head)
        return comments, node

    def handle_inline(self, node):
        if self.is_if_multiline(node):
            comments, node = self.flatten_inline_if(node)
        else:
            comments = []
        if self.is_shorter_than_limit(node):  # TODO ignore comments?
            return (*comments, node)
        indent = node.header.tokens[0]
        separator = self.formatting_config.separator
        assign_tokens = node.header.get_tokens(Token.ASSIGN)
        assign = [*self.join_on_separator(assign_tokens, Token(Token.SEPARATOR, separator))]
        else_present = False
        branches = []
        while node:
            new_comments, if_block, else_found = self.handle_inline_if_create(node, indent.value, assign)
            else_present = else_present or else_found
            comments += new_comments
            branches.append(if_block)
            node = node.orelse
        if not else_present and assign_tokens:
            header = ElseHeader.from_params(indent=indent.value)
            keyword = self.create_keyword_for_inline(
                [
                    Token(Token.KEYWORD, "Set Variable"),
                    *[Token(Token.ARGUMENT, "${None}") for _ in range(len(assign_tokens))],
                ],
                indent.value,
                assign,
            )
            branches.append(If(header=header, body=[keyword]))
        if_block = head = branches[0]
        for branch in branches[1:]:
            head.orelse = branch
            head = head.orelse
        if_block.end = End([indent, Token(Token.END), Token(Token.EOL)])
        return (*comments, if_block)

    def handle_inline_if_create(self, node, indent, assign):
        comments = self.collect_comments_from_if(indent, node)
        body = [self.create_keyword_for_inline(node.body[0].data_tokens, indent, assign)]
        else_found = False
        if isinstance(node.header, InlineIfHeader):
            header = IfHeader.from_params(
                condition=node.condition, indent=indent, separator=self.formatting_config.separator
            )
        elif isinstance(node.header, ElseIfHeader):
            header = ElseIfHeader.from_params(
                condition=node.condition, indent=indent, separator=self.formatting_config.separator
            )
        else:
            header = ElseHeader.from_params(indent=indent)
            else_found = True
        return comments, If(header=header, body=body), else_found



import logging
from typing import Mapping, Any, List, Optional, Dict

from robocorp_ls_core.basic import implements
from robocorp_ls_core.client_base import LanguageServerClientBase
from robocorp_ls_core.protocols import (
    ILanguageServerClient,
    IIdMessageMatcher,
)
from robocorp_ls_core.lsp import (
    CodeLensTypedDict,
    CompletionsResponseTypedDict,
    CompletionItemTypedDict,
    CompletionResolveResponseTypedDict,
)


log = logging.getLogger(__name__)


class LanguageServerClient(LanguageServerClientBase):
    pid: Optional[int] = None

    DEFAULT_TIMEOUT: Optional[int] = None

    def __init__(self, *args, **kwargs):
        from robocorp_ls_core.unittest_tools.fixtures import TIMEOUT

        LanguageServerClientBase.__init__(self, *args, **kwargs)

        self.DEFAULT_TIMEOUT = TIMEOUT

    @implements(ILanguageServerClient.settings)
    def settings(self, settings: Dict):
        self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "workspace/didChangeConfiguration",
                "params": settings,
            }
        )

    @implements(ILanguageServerClient.initialize)
    def initialize(
        self, root_path: str, msg_id=None, process_id=None, initialization_options=None
    ):
        from robocorp_ls_core.uris import from_fs_path

        root_uri = from_fs_path(root_path)

        msg_id = msg_id if msg_id is not None else self.next_id()
        msg = self.request(
            {
                "jsonrpc": "2.0",
                "id": msg_id,
                "method": "initialize",
                "params": {
                    "processId": process_id,
                    "rootPath": root_path,
                    "rootUri": root_uri,
                    "initializationOptions": initialization_options,
                    "capabilities": {
                        "workspace": {
                            "applyEdit": True,
                            "didChangeConfiguration": {"dynamicRegistration": True},
                            "didChangeWatchedFiles": {"dynamicRegistration": True},
                            "symbol": {"dynamicRegistration": True},
                            "executeCommand": {"dynamicRegistration": True},
                        },
                        "textDocument": {
                            "synchronization": {
                                "dynamicRegistration": True,
                                "willSave": True,
                                "willSaveWaitUntil": True,
                                "didSave": True,
                            },
                            "completion": {
                                "dynamicRegistration": True,
                                "completionItem": {
                                    "snippetSupport": True,
                                    "commitCharactersSupport": True,
                                },
                            },
                            "hover": {"dynamicRegistration": True},
                            "signatureHelp": {"dynamicRegistration": True},
                            "definition": {"dynamicRegistration": True},
                            "references": {"dynamicRegistration": True},
                            "documentHighlight": {"dynamicRegistration": True},
                            "documentSymbol": {"dynamicRegistration": True},
                            "codeAction": {"dynamicRegistration": True},
                            "codeLens": {"dynamicRegistration": True},
                            "formatting": {"dynamicRegistration": True},
                            "rangeFormatting": {"dynamicRegistration": True},
                            "onTypeFormatting": {"dynamicRegistration": True},
                            "rename": {"dynamicRegistration": True},
                            "documentLink": {"dynamicRegistration": True},
                        },
                    },
                    "trace": "off",
                },
            }
        )

        assert "capabilities" in msg["result"]
        return msg

    @implements(ILanguageServerClient.change_workspace_folders)
    def change_workspace_folders(
        self, added_folders: List[str], removed_folders: List[str]
    ) -> None:
        from robocorp_ls_core import uris
        import os.path

        added_folders_uri_name = [
            {"uri": uris.from_fs_path(s), "name": os.path.basename(s)}
            for s in added_folders
        ]
        removed_folders_uri_name = [
            {"uri": uris.from_fs_path(s), "name": os.path.basename(s)}
            for s in removed_folders
        ]
        self.request(
            {
                "jsonrpc": "2.0",
                "method": "workspace/didChangeWorkspaceFolders",
                "id": self.next_id(),
                "params": {
                    "event": {
                        "added": added_folders_uri_name,
                        "removed": removed_folders_uri_name,
                    }
                },
            }
        )

    @implements(ILanguageServerClient.open_doc)
    def open_doc(self, uri: str, version: int = 1, text: Optional[str] = None):
        self.request(
            {
                "jsonrpc": "2.0",
                "method": "textDocument/didOpen",
                "id": self.next_id(),
                "params": {
                    "textDocument": {
                        "uri": uri,
                        "languageId": "robotframework",
                        "version": version,
                        "text": text,
                    }
                },
            }
        )

    @implements(ILanguageServerClient.close_doc)
    def close_doc(self, uri: str):
        self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/didClose",
                "params": {"textDocument": {"uri": uri}},
            }
        )

    @implements(ILanguageServerClient.hover)
    def hover(self, uri: str, line: int, col: int):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/hover",
                "params": {
                    "textDocument": {"uri": uri},
                    "position": {"line": line, "character": col},
                },
            }
        )

    @implements(ILanguageServerClient.change_doc)
    def change_doc(self, uri: str, version: int, text: str):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/didChange",
                "params": {
                    "textDocument": {"uri": uri, "version": version},
                    "contentChanges": [{"range": None, "rangeLength": 0, "text": text}],
                },
            }
        )

    def _build_completions_request(self, uri: str, line: int, col: int):
        return {
            "jsonrpc": "2.0",
            "id": self.next_id(),
            "method": "textDocument/completion",
            "params": {
                "textDocument": {"uri": uri},
                "position": {"line": line, "character": col},
            },
        }

    @implements(ILanguageServerClient.get_completions)
    def get_completions(
        self, uri: str, line: int, col: int
    ) -> CompletionsResponseTypedDict:
        return self.request(self._build_completions_request(uri, line, col))

    @implements(ILanguageServerClient.get_completions_async)
    def get_completions_async(
        self, uri: str, line: int, col: int
    ) -> Optional[IIdMessageMatcher[CompletionsResponseTypedDict]]:
        return self.request_async(self._build_completions_request(uri, line, col))

    @implements(ILanguageServerClient.request_resolve_completion)
    def request_resolve_completion(
        self, completion_item: CompletionItemTypedDict
    ) -> CompletionResolveResponseTypedDict:
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "completionItem/resolve",
                "params": completion_item,
            }
        )

    @implements(ILanguageServerClient.request_source_format)
    def request_source_format(self, uri: str):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/formatting",
                "params": {"textDocument": {"uri": uri}},
            }
        )

    @implements(ILanguageServerClient.request_signature_help)
    def request_signature_help(self, uri, line, col):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/signatureHelp",
                "params": {
                    "textDocument": {"uri": uri},
                    "position": {"line": line, "character": col},
                },
            }
        )

    @implements(ILanguageServerClient.request_hover)
    def request_hover(self, uri, line, col):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/hover",
                "params": {
                    "textDocument": {"uri": uri},
                    "position": {"line": line, "character": col},
                },
            }
        )

    @implements(ILanguageServerClient.request_text_document_highlight)
    def request_text_document_highlight(self, uri, line, col):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/documentHighlight",
                "params": {
                    "textDocument": {"uri": uri},
                    "position": {"line": line, "character": col},
                },
            }
        )

    @implements(ILanguageServerClient.request_references)
    def request_references(self, uri, line, col, include_declaration):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/references",
                "params": {
                    "textDocument": {"uri": uri},
                    "position": {"line": line, "character": col},
                    "context": {
                        "includeDeclaration": include_declaration,
                    },
                },
            }
        )

    @implements(ILanguageServerClient.request_folding_range)
    def request_folding_range(self, uri):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/foldingRange",
                "params": {"textDocument": {"uri": uri}},
            }
        )

    @implements(ILanguageServerClient.request_code_lens)
    def request_code_lens(self, uri):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/codeLens",
                "params": {"textDocument": {"uri": uri}},
            }
        )

    @implements(ILanguageServerClient.request_resolve_code_lens)
    def request_resolve_code_lens(self, code_lens: CodeLensTypedDict):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "codeLens/resolve",
                "params": code_lens,
            }
        )

    @implements(ILanguageServerClient.request_document_symbol)
    def request_document_symbol(self, uri):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/document_symbol",
                "params": {"textDocument": {"uri": uri}},
            }
        )

    @implements(ILanguageServerClient.request_workspace_symbols)
    def request_workspace_symbols(self, query: Optional[str] = None):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "workspace/symbol",
                "params": {"query": query},
            }
        )

    def request_cancel(self, message_id) -> None:
        """
        Requests that some processing is cancelled.
        """
        self.write(
            {
                "jsonrpc": "2.0",
                "method": "$/cancelRequest",
                "params": {"id": message_id},
            }
        )

    @implements(ILanguageServerClient.find_definitions)
    def find_definitions(self, uri, line: int, col: int):
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "textDocument/definition",
                "params": {
                    "textDocument": {"uri": uri},
                    "position": {"line": line, "character": col},
                },
            }
        )

    @implements(ILanguageServerClient.execute_command)
    def execute_command(self, command: str, arguments: list) -> Mapping[str, Any]:
        return self.request(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "workspace/executeCommand",
                "params": {"command": command, "arguments": arguments},
            }
        )

    @implements(ILanguageServerClient.execute_command_async)
    def execute_command_async(
        self, command: str, arguments: list
    ) -> Optional[IIdMessageMatcher]:
        return self.request_async(
            {
                "jsonrpc": "2.0",
                "id": self.next_id(),
                "method": "workspace/executeCommand",
                "params": {"command": command, "arguments": arguments},
            }
        )

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: ILanguageServerClient = check_implements(self)



import multiprocessing
import sys


def run(name):
    print("argument: ", name)  # break 1 here


if __name__ == '__main__':
    if sys.version_info[0] >= 3 and sys.platform != 'win32':
        multiprocessing.set_start_method('fork')
    p = multiprocessing.Process(target=run, args=("argument to run method",))
    p.start()
    print('TEST SUCEEDED!')  # break 2 here
    p.join()



r'''
Full setup, used to distribute the debugger backend to PyPi.

Note that this is mostly so that users can do:

pip install pydevd

in a machine for doing remote-debugging, as a local installation with the IDE should have
everything already distributed.

Reference on wheels:
https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
http://lucumr.pocoo.org/2014/1/27/python-on-wheels/

Another (no wheels): https://jamie.curle.io/blog/my-first-experience-adding-package-pypi/

See:

build_tools\pydevd_release_process.txt

for release process.
'''

from setuptools import setup
from setuptools.dist import Distribution
from distutils.extension import Extension
import os


class BinaryDistribution(Distribution):

    def is_pure(self):
        return False


data_files = []


def accept_file(f):
    f = f.lower()
    for ext in '.py .dll .so .dylib .txt .cpp .h .bat .c .sh .md .txt'.split():
        if f.endswith(ext):
            return True

    return f in ['readme', 'makefile']


data_files.append(('pydevd_attach_to_process', [os.path.join('pydevd_attach_to_process', f) for f in os.listdir('pydevd_attach_to_process') if accept_file(f)]))
for root, dirs, files in os.walk("pydevd_attach_to_process"):
    for d in dirs:
        data_files.append((os.path.join(root, d), [os.path.join(root, d, f) for f in os.listdir(os.path.join(root, d)) if accept_file(f)]))

import pydevd
version = pydevd.__version__

args = dict(
    name='pydevd',
    version=version,
    description='PyDev.Debugger (used in PyDev, PyCharm and VSCode Python)',
    author='Fabio Zadrozny and others',
    url='https://github.com/fabioz/PyDev.Debugger/',
    license='EPL (Eclipse Public License)',
    packages=[
        '_pydev_bundle',
        '_pydev_bundle.fsnotify',
        '_pydev_runfiles',
        '_pydevd_bundle',
        '_pydevd_bundle._debug_adapter',
        '_pydevd_bundle.pydevd_concurrency_analyser',
        '_pydevd_frame_eval',
        '_pydevd_frame_eval.vendored',
        '_pydevd_frame_eval.vendored.bytecode',
        'pydev_ipython',

        # 'pydev_sitecustomize', -- Not actually a package (not added)

        'pydevd_attach_to_process',

        'pydevd_plugins',
        'pydevd_plugins.extensions',
        'pydevd_plugins.extensions.types',
    ],
    py_modules=[
        # 'interpreterInfo', -- Not needed for debugger
        # 'pycompletionserver', -- Not needed for debugger
        'pydev_app_engine_debug_startup',
        # 'pydev_coverage', -- Not needed for debugger
        # 'pydev_pysrc', -- Not needed for debugger
        'pydev_run_in_console',
        'pydevconsole',
        'pydevd_file_utils',
        'pydevd',
        'pydevd_tracing',
        # 'runfiles', -- Not needed for debugger
        'setup_pydevd_cython',  # Distributed to clients. See: https://github.com/fabioz/PyDev.Debugger/issues/102
        # 'setup', -- Should not be included as a module
    ],
    classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Console',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: Eclipse Public License 1.0 (EPL-1.0)',

        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Debuggers',
    ],
    entry_points={
        'console_scripts':[
            'pydevd = pydevd:main',
        ],
    },
    data_files=data_files,
    keywords=['pydev', 'pydevd', 'pydev.debugger'],
    include_package_data=True,
    zip_safe=False,
)

import sys
try:
    extra_compile_args = []
    extra_link_args = []

    if 'linux' in sys.platform:
        # Enabling -flto brings executable from 4MB to 0.56MB and -Os to 0.41MB
        # Profiling shows an execution around 3-5% slower with -Os vs -O3,
        # so, kept only -flto.
        extra_compile_args = ["-flto", "-O3"]
        extra_link_args = extra_compile_args[:]

        # Note: also experimented with profile-guided optimization. The executable
        # size became a bit smaller (from 0.56MB to 0.5MB) but this would add an
        # extra step to run the debugger to obtain the optimizations
        # so, skipped it for now (note: the actual benchmarks time was in the
        # margin of a 0-1% improvement, which is probably not worth it for
        # speed increments).
        # extra_compile_args = ["-flto", "-fprofile-generate"]
        # ... Run benchmarks ...
        # extra_compile_args = ["-flto", "-fprofile-use", "-fprofile-correction"]
    elif 'win32' in sys.platform:
        pass
        # uncomment to generate pdbs for visual studio.
        # extra_compile_args=["-Zi", "/Od"]
        # extra_link_args=["-debug"]

    kwargs = {}
    if extra_link_args:
        kwargs['extra_link_args'] = extra_link_args
    if extra_compile_args:
        kwargs['extra_compile_args'] = extra_compile_args

    args_with_binaries = args.copy()
    args_with_binaries.update(dict(
        distclass=BinaryDistribution,
        ext_modules=[
            # In this setup, don't even try to compile with cython, just go with the .c file which should've
            # been properly generated from a tested version.
            Extension(
                '_pydevd_bundle.pydevd_cython',
                ["_pydevd_bundle/pydevd_cython.c", ],
                define_macros=[('Py_BUILD_CORE_MODULE', '1')],
                **kwargs
            )
        ]
    ))
    setup(**args_with_binaries)
except:
    # Compile failed: just setup without compiling cython deps.
    setup(**args)
    sys.stdout.write('Plain-python version of pydevd installed (cython speedups not available).\n')



def check():
    from collections import OrderedDict
    import sys
    # On 3.6 onwards, use a regular dict.
    odict = OrderedDict() if sys.version_info[:2] < (3, 6) else {}
    odict[4] = 'first'
    odict[3] = 'second'
    odict[2] = 'last'
    print('break here')


if __name__ == '__main__':
    check()
print('TEST SUCEEDED')



"""Built-in template tests used with the ``is`` operator."""
import operator
import typing as t
from collections import abc
from numbers import Number

from .runtime import Undefined
from .utils import pass_environment

if t.TYPE_CHECKING:
    from .environment import Environment


def test_odd(value: int) -> bool:
    """Return true if the variable is odd."""
    return value % 2 == 1


def test_even(value: int) -> bool:
    """Return true if the variable is even."""
    return value % 2 == 0


def test_divisibleby(value: int, num: int) -> bool:
    """Check if a variable is divisible by a number."""
    return value % num == 0


def test_defined(value: t.Any) -> bool:
    """Return true if the variable is defined:

    .. sourcecode:: jinja

        {% if variable is defined %}
            value of variable: {{ variable }}
        {% else %}
            variable is not defined
        {% endif %}

    See the :func:`default` filter for a simple way to set undefined
    variables.
    """
    return not isinstance(value, Undefined)


def test_undefined(value: t.Any) -> bool:
    """Like :func:`defined` but the other way round."""
    return isinstance(value, Undefined)


@pass_environment
def test_filter(env: "Environment", value: str) -> bool:
    """Check if a filter exists by name. Useful if a filter may be
    optionally available.

    .. code-block:: jinja

        {% if 'markdown' is filter %}
            {{ value | markdown }}
        {% else %}
            {{ value }}
        {% endif %}

    .. versionadded:: 3.0
    """
    return value in env.filters


@pass_environment
def test_test(env: "Environment", value: str) -> bool:
    """Check if a test exists by name. Useful if a test may be
    optionally available.

    .. code-block:: jinja

        {% if 'loud' is test %}
            {% if value is loud %}
                {{ value|upper }}
            {% else %}
                {{ value|lower }}
            {% endif %}
        {% else %}
            {{ value }}
        {% endif %}

    .. versionadded:: 3.0
    """
    return value in env.tests


def test_none(value: t.Any) -> bool:
    """Return true if the variable is none."""
    return value is None


def test_boolean(value: t.Any) -> bool:
    """Return true if the object is a boolean value.

    .. versionadded:: 2.11
    """
    return value is True or value is False


def test_false(value: t.Any) -> bool:
    """Return true if the object is False.

    .. versionadded:: 2.11
    """
    return value is False


def test_true(value: t.Any) -> bool:
    """Return true if the object is True.

    .. versionadded:: 2.11
    """
    return value is True


# NOTE: The existing 'number' test matches booleans and floats
def test_integer(value: t.Any) -> bool:
    """Return true if the object is an integer.

    .. versionadded:: 2.11
    """
    return isinstance(value, int) and value is not True and value is not False


# NOTE: The existing 'number' test matches booleans and integers
def test_float(value: t.Any) -> bool:
    """Return true if the object is a float.

    .. versionadded:: 2.11
    """
    return isinstance(value, float)


def test_lower(value: str) -> bool:
    """Return true if the variable is lowercased."""
    return str(value).islower()


def test_upper(value: str) -> bool:
    """Return true if the variable is uppercased."""
    return str(value).isupper()


def test_string(value: t.Any) -> bool:
    """Return true if the object is a string."""
    return isinstance(value, str)


def test_mapping(value: t.Any) -> bool:
    """Return true if the object is a mapping (dict etc.).

    .. versionadded:: 2.6
    """
    return isinstance(value, abc.Mapping)


def test_number(value: t.Any) -> bool:
    """Return true if the variable is a number."""
    return isinstance(value, Number)


def test_sequence(value: t.Any) -> bool:
    """Return true if the variable is a sequence. Sequences are variables
    that are iterable.
    """
    try:
        len(value)
        value.__getitem__
    except Exception:
        return False

    return True


def test_sameas(value: t.Any, other: t.Any) -> bool:
    """Check if an object points to the same memory address than another
    object:

    .. sourcecode:: jinja

        {% if foo.attribute is sameas false %}
            the foo attribute really is the `False` singleton
        {% endif %}
    """
    return value is other


def test_iterable(value: t.Any) -> bool:
    """Check if it's possible to iterate over an object."""
    try:
        iter(value)
    except TypeError:
        return False

    return True


def test_escaped(value: t.Any) -> bool:
    """Check if the value is escaped."""
    return hasattr(value, "__html__")


def test_in(value: t.Any, seq: t.Container) -> bool:
    """Check if value is in seq.

    .. versionadded:: 2.10
    """
    return value in seq


TESTS = {
    "odd": test_odd,
    "even": test_even,
    "divisibleby": test_divisibleby,
    "defined": test_defined,
    "undefined": test_undefined,
    "filter": test_filter,
    "test": test_test,
    "none": test_none,
    "boolean": test_boolean,
    "false": test_false,
    "true": test_true,
    "integer": test_integer,
    "float": test_float,
    "lower": test_lower,
    "upper": test_upper,
    "string": test_string,
    "mapping": test_mapping,
    "number": test_number,
    "sequence": test_sequence,
    "iterable": test_iterable,
    "callable": callable,
    "sameas": test_sameas,
    "escaped": test_escaped,
    "in": test_in,
    "==": operator.eq,
    "eq": operator.eq,
    "equalto": operator.eq,
    "!=": operator.ne,
    "ne": operator.ne,
    ">": operator.gt,
    "gt": operator.gt,
    "greaterthan": operator.gt,
    "ge": operator.ge,
    ">=": operator.ge,
    "<": operator.lt,
    "lt": operator.lt,
    "lessthan": operator.lt,
    "<=": operator.le,
    "le": operator.le,
}



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

from __future__ import absolute_import, division, print_function, unicode_literals

import contextlib
from importlib import import_module
import os
import sys
from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)

VENDORED_ROOT = os.path.dirname(os.path.abspath(__file__))


def project_root(project):
    """Return the path to the root dir of the vendored project.

    If "project" is an empty string then the path prefix for vendored
    projects (e.g. "robotframework_debug_adapter/_vendored/") will be returned.
    """
    if not project:
        project = ""
    return os.path.join(VENDORED_ROOT, project)


@contextlib.contextmanager
def vendored(project, root=None):
    """A context manager under which the vendored project will be imported."""
    if root is None:
        root = project_root(project)
    # Add the vendored project directory, so that it gets tried first.
    sys.path.insert(0, root)
    try:
        yield root
    finally:
        sys.path.remove(root)


def preimport(project, modules, **kwargs):
    """Import each of the named modules out of the vendored project."""
    with vendored(project, **kwargs):
        for name in modules:
            import_module(name)


try:
    import pydevd  # noqa
except ImportError:
    pydevd_available = False
else:
    pydevd_available = True


if not pydevd_available:
    # Constants must be set before importing any other pydevd module
    # # due to heavy use of "from" in them.
    with vendored("vendored_pydevd"):
        try:
            pydevd_constants = import_module("_pydevd_bundle.pydevd_constants")
        except ImportError as e:
            contents = os.listdir(VENDORED_ROOT)

            for c in contents[:]:
                if os.path.isdir(c):
                    contents.append(f"{c}/{os.listdir(c)}")
                else:
                    contents.append(c)

            s = "\n".join(contents)

            msg = f"Vendored root: {VENDORED_ROOT} -- contents:\n{s}"
            raise ImportError(msg) from e

    # Now make sure all the top-level modules and packages in pydevd are
    # loaded.  Any pydevd modules that aren't loaded at this point, will
    # be loaded using their parent package's __path__ (i.e. one of the
    # following).
    preimport(
        "vendored_pydevd",
        [
            "_pydev_bundle",
            "_pydev_runfiles",
            "_pydevd_bundle",
            "_pydevd_frame_eval",
            "pydev_ipython",
            "pydevd_plugins",
            "pydevd",
        ],
    )

import pydevd  # noqa

log.info("Vendored root: %s\npydevd: %s", VENDORED_ROOT, pydevd)

# Ensure that pydevd uses JSON protocol by default.
from _pydevd_bundle import pydevd_constants
from _pydevd_bundle import pydevd_defaults

pydevd_defaults.PydevdCustomization.DEFAULT_PROTOCOL = (
    pydevd_constants.HTTP_JSON_PROTOCOL
)


from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
    BaseSchema as RobotSchema,
)
from _pydevd_bundle._debug_adapter.pydevd_base_schema import BaseSchema as PyDevdSchema

PyDevdSchema._obj_id_to_dap_id = RobotSchema._obj_id_to_dap_id
PyDevdSchema._dap_id_to_obj_id = RobotSchema._dap_id_to_obj_id
PyDevdSchema._next_dap_id = RobotSchema._next_dap_id



import sys
from typing import (
    Iterator,
    Optional,
    List,
    Tuple,
    Any,
    Union,
    Hashable,
    Callable,
    Dict,
    Iterable,
)

import ast as ast_module
from robocorp_ls_core.lsp import Error, RangeTypedDict, PositionTypedDict
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    TokenInfo,
    NodeInfo,
    KeywordUsageInfo,
    ILibraryImportNode,
    IRobotToken,
    INode,
    IRobotVariableMatch,
    VarTokenInfo,
    IKeywordArg,
)
from robotframework_ls.impl.text_utilities import normalize_robot_name
from robocorp_ls_core.basic import isinstance_name
from robotframework_ls.impl.keywords_in_args import (
    KEYWORD_NAME_TO_KEYWORD_INDEX,
    KEYWORD_NAME_TO_CONDITION_INDEX,
)
import functools
import weakref
import threading
import typing


log = get_logger(__name__)


class _NodesProviderVisitor(ast_module.NodeVisitor):
    def __init__(self, on_node=lambda node: None):
        ast_module.NodeVisitor.__init__(self)
        self._stack = []
        self.on_node = on_node

    def generic_visit(self, node):
        self._stack.append(node)
        self.on_node(self._stack, node)
        ast_module.NodeVisitor.generic_visit(self, node)
        self._stack.pop()


class _PrinterVisitor(ast_module.NodeVisitor):
    def __init__(self, stream):
        ast_module.NodeVisitor.__init__(self)
        self._level = 0
        self._stream = stream

    def _replace_spacing(self, txt):
        curr_len = len(txt)
        delta = 80 - curr_len
        return txt.replace("*SPACING*", " " * delta)

    def generic_visit(self, node):
        # Note: prints line and col offsets 0-based (even if the ast is 1-based for
        # lines and 0-based for columns).
        self._level += 1
        try:
            indent = "  " * self._level
            node_lineno = node.lineno
            if node_lineno != -1:
                # Make 0-based
                node_lineno -= 1
            node_end_lineno = node.end_lineno
            if node_end_lineno != -1:
                # Make 0-based
                node_end_lineno -= 1
            self._stream.write(
                self._replace_spacing(
                    "%s%s *SPACING* (%s, %s) -> (%s, %s)\n"
                    % (
                        indent,
                        node.__class__.__name__,
                        node_lineno,
                        node.col_offset,
                        node_end_lineno,
                        node.end_col_offset,
                    )
                )
            )
            tokens = getattr(node, "tokens", [])
            for token in tokens:

                token_lineno = token.lineno
                if token_lineno != -1:
                    # Make 0-based
                    token_lineno -= 1

                self._stream.write(
                    self._replace_spacing(
                        "%s- %s, '%s' *SPACING* (%s, %s->%s)\n"
                        % (
                            indent,
                            token.type,
                            token.value.replace("\n", "\\n").replace("\r", "\\r"),
                            token_lineno,
                            token.col_offset,
                            token.end_col_offset,
                        )
                    )
                )

            ast_module.NodeVisitor.generic_visit(self, node)
        finally:
            self._level -= 1


MAX_ERRORS = 100


class _AbstractIndexer:
    def iter_indexed(self, clsname):
        pass

    @property
    def ast(self):
        return self._weak_ast()


class _FullIndexer(_AbstractIndexer):
    def __init__(self, weak_ast: "weakref.ref[ast_module.AST]"):
        self._weak_ast = weak_ast
        self._lock = threading.Lock()
        self._name_to_node_info_lst: Dict[str, List[NodeInfo]] = {}
        self._indexed_full = False

    def _index(self):
        with self._lock:
            if self._indexed_full:
                return

            ast = self._weak_ast()
            if ast is None:
                raise RuntimeError("AST already garbage collected.")

            for stack, node in _iter_nodes(ast):
                lst = self._name_to_node_info_lst.get(node.__class__.__name__)
                if lst is None:
                    lst = self._name_to_node_info_lst[node.__class__.__name__] = []

                lst.append(NodeInfo(tuple(stack), node))
            self._indexed_full = True

    def iter_indexed(self, clsname: str) -> Iterator[NodeInfo]:
        if not self._indexed_full:
            self._index()

        yield from iter(self._name_to_node_info_lst.get(clsname, ()))


class _SectionIndexer(_AbstractIndexer):
    """
    This is a bit smarter in that it can index only the parts we're interested
    in (so, to get the LibraryImport it won't iterate over the keywords to
    do the indexing).
    """

    INNER_INSIDE_TOP_LEVEL = {
        "LibraryImport": "SettingSection",
        "ResourceImport": "SettingSection",
        "VariablesImport": "SettingSection",
        "SuiteSetup": "SettingSection",
        "SuiteTeardown": "SettingSection",
        "TestTemplate": "SettingSection",
        # Not settings:
        "Keyword": "KeywordSection",
        "TestCase": "TestCaseSection",
        "Variable": "VariableSection",
    }

    TOP_LEVEL = {
        "SettingSection",
        "VariableSection",
        "TestCaseSection",
        "KeywordSection",
        "CommentSection",
    }

    def __init__(self, weak_ast):
        self._weak_ast = weak_ast
        self._lock = threading.Lock()
        self._first_level_name_to_node_info_lst: Dict[str, List[NodeInfo]] = {}

        # We always start by indexing the first level in this case (to get the sections
        # such as 'CommentSection', 'SettingSection', etc), which should be fast.

        ast = self._weak_ast()
        if ast is None:
            raise RuntimeError("AST already garbage collected.")

        for stack, node in _iter_nodes(ast, recursive=False):
            lst = self._first_level_name_to_node_info_lst.get(node.__class__.__name__)
            if lst is None:
                lst = self._first_level_name_to_node_info_lst[
                    node.__class__.__name__
                ] = []

            lst.append(NodeInfo(tuple(stack), node))

    def iter_indexed(self, clsname: str) -> Iterator[NodeInfo]:
        top_level = self.INNER_INSIDE_TOP_LEVEL.get(clsname)
        if top_level is not None:
            lst = self._first_level_name_to_node_info_lst.get(top_level)
            if lst is not None:
                for node_info in lst:
                    indexer = _obtain_ast_indexer(node_info.node)
                    yield from indexer.iter_indexed(clsname)
        else:
            if clsname in self.TOP_LEVEL:
                yield from iter(
                    self._first_level_name_to_node_info_lst.get(clsname, ())
                )
            else:
                # i.e.: We don't know what we should be getting, so, just check
                # everything...
                for lst in self._first_level_name_to_node_info_lst.values():
                    for node_info in lst:
                        indexer = _obtain_ast_indexer(node_info.node)
                        yield from indexer.iter_indexed(clsname)


class _ASTIndexer(_AbstractIndexer):
    def __init__(self, ast: ast_module.AST):
        self._weak_ast = weakref.ref(ast)
        self._is_root = ast.__class__.__name__ == "File"

        self._indexer: _AbstractIndexer
        if self._is_root:
            # Cache by sections
            self._indexer = _SectionIndexer(self._weak_ast)
        else:
            # Always cache fully
            self._indexer = _FullIndexer(self._weak_ast)

        self._additional_caches: Dict[Hashable, Tuple[Any, ...]] = {}

    def iter_cached(
        self, cache_key: Hashable, compute: Callable, *args
    ) -> Iterator[Any]:
        try:
            cached = self._additional_caches[cache_key]
        except KeyError:
            cached = tuple(compute(self, *args))
            self._additional_caches[cache_key] = cached

        yield from iter(cached)

    def iter_indexed(self, clsname: str) -> Iterator[NodeInfo]:
        return self._indexer.iter_indexed(clsname)


def _get_errors_from_tokens(node):
    for token in node.tokens:
        if token.type in (token.ERROR, token.FATAL_ERROR):
            start = (token.lineno - 1, token.col_offset)
            end = (token.lineno - 1, token.end_col_offset)
            error = Error(token.error, start, end)
            yield error


def _obtain_ast_indexer(ast):
    try:
        indexer = ast.__ast_indexer__
    except:
        indexer = ast.__ast_indexer__ = _ASTIndexer(ast)
    return indexer


def _convert_ast_to_indexer(func):
    @functools.wraps(func)
    def new_func(ast, *args, **kwargs):
        try:
            indexer = ast.__ast_indexer__
        except:
            indexer = ast.__ast_indexer__ = _ASTIndexer(ast)

        return func(indexer, *args, **kwargs)

    return new_func


def collect_errors(node) -> List[Error]:
    errors = []

    use_errors_attribute = "errors" in node.__class__._attributes

    for _stack, node in _iter_nodes(node, recursive=True):
        if node.__class__.__name__ == "Error":
            errors.extend(_get_errors_from_tokens(node))

        elif use_errors_attribute:
            node_errors = getattr(node, "errors", ())
            if node_errors:
                for error in node_errors:
                    errors.append(create_error_from_node(node, error, tokens=[node]))

        if len(errors) >= MAX_ERRORS:
            break

    return errors


def create_error_from_node(node, msg, tokens=None) -> Error:
    if tokens is None:
        tokens = node.tokens

    if not tokens:
        log.info("No tokens found when visiting: %s.", node.__class__)
        start = (0, 0)
        end = (0, 0)
    else:
        # line is 1-based and col is 0-based (make both 0-based for the error).
        start = (tokens[0].lineno - 1, tokens[0].col_offset)
        end = (tokens[-1].lineno - 1, tokens[-1].end_col_offset)

    error = Error(msg, start, end)
    return error


def print_ast(node, stream=None):
    if stream is None:
        stream = sys.stderr
    errors_visitor = _PrinterVisitor(stream)
    errors_visitor.visit(node)


def find_section(node, line: int) -> Optional[INode]:
    """
    :param line:
        0-based
    """
    last_section = None
    for section in node.sections:
        # section.lineno is 1-based.
        if (section.lineno - 1) <= line:
            last_section = section

        else:
            return last_section

    return last_section


if typing.TYPE_CHECKING:
    # The INode has Robot Framework specific methods, but at runtime
    # we can just check the actual ast class.
    from typing import runtime_checkable, Protocol

    @runtime_checkable
    class _AST_CLASS(INode, Protocol):
        pass


else:
    # We know that the AST we're dealing with is the INode.
    # We can't use runtime_checkable on Python 3.7 though.
    _AST_CLASS = ast_module.AST


def _iter_nodes(
    node, internal_stack: Optional[List[INode]] = None, recursive=True
) -> Iterator[Tuple[List[INode], INode]]:
    """
    :note: the yielded stack is actually always the same (mutable) list, so,
    clients that want to return it somewhere else should create a copy.
    """
    stack: List[INode]
    if internal_stack is None:
        stack = []
    else:
        stack = internal_stack

    if recursive:
        for _field, value in ast_module.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, _AST_CLASS):
                        yield stack, item
                        stack.append(item)
                        yield from _iter_nodes(item, stack, recursive=True)
                        stack.pop()

            elif isinstance(value, _AST_CLASS):
                yield stack, value
                stack.append(value)

                yield from _iter_nodes(value, stack, recursive=True)

                stack.pop()
    else:
        # Not recursive
        for _field, value in ast_module.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, _AST_CLASS):
                        yield stack, item

            elif isinstance(value, _AST_CLASS):
                yield stack, value


def iter_all_nodes_recursive(node: INode) -> Iterator[Tuple[List[INode], INode]]:
    """
    This function will iterate over all the nodes. Use only if there's no
    other way to implement it as iterating over all the nodes is slow...
    """
    yield from _iter_nodes(node)


def _iter_nodes_filtered_not_recursive(
    ast, accept_class: Union[Tuple[str, ...], str]
) -> Iterator[Tuple[list, Any]]:
    if not isinstance(accept_class, (list, tuple, set)):
        accept_class = (accept_class,)
    for stack, node in _iter_nodes(ast, recursive=False):
        if node.__class__.__name__ in accept_class:
            yield stack, node


def find_token(section, line, col) -> Optional[TokenInfo]:
    """
    :param section:
        The result from find_section(line, col), to pre-filter the nodes we may match.
    """
    for stack, node in _iter_nodes(section):
        try:
            tokens = node.tokens
        except AttributeError:
            continue

        if not tokens:
            continue

        if (tokens[-1].lineno - 1) < line:
            # i.e.: if the last node in the token is still before the
            # line we're searching, keep on going
            continue

        last_token = None
        for token in tokens:
            lineno = token.lineno - 1
            if lineno != line:
                continue

            if token.type == token.SEPARATOR:
                # For separator tokens, it must be entirely within the section
                # i.e.: if it's in the boundary for a word, we want the word,
                # not the separator.
                if token.col_offset < col < token.end_col_offset:
                    return TokenInfo(tuple(stack), node, token)

            elif token.type == token.EOL:
                # A trailing whitespace after a keyword should be part of
                # the keyword, not EOL.
                if token.col_offset <= col <= token.end_col_offset:
                    diff = col - token.col_offset
                    if last_token is not None and not token.value.strip():
                        eol_contents = token.value[:diff]
                        if len(eol_contents) <= 1:
                            token = _append_eol_to_prev_token(last_token, eol_contents)

                    return TokenInfo(tuple(stack), node, token)

            else:
                if token.col_offset <= col <= token.end_col_offset:
                    return TokenInfo(tuple(stack), node, token)

            last_token = token

    return None


def _find_subvar(stack, node, initial_token, col) -> Optional[VarTokenInfo]:
    for token, var_identifier in _tokenize_subvars(initial_token):
        if token.col_offset <= col <= token.end_col_offset:
            return VarTokenInfo(stack, node, token, var_identifier)

    from robotframework_ls.impl import robot_constants

    for p in robot_constants.VARIABLE_PREFIXES:
        if initial_token.value.startswith(p + "{"):
            if initial_token.value.endswith("}"):
                var_token = copy_token_with_subpart(initial_token, 2, -1)
            else:
                var_token = copy_token_with_subpart(
                    initial_token, 2, len(initial_token.value)
                )

            var_identifier = p
            return VarTokenInfo(stack, node, var_token, var_identifier)
    return None


def find_variable(section, line, col) -> Optional[VarTokenInfo]:
    """
    Finds the current variable token. Note that it won't include '{' nor '}'.
    The token may also be an empty token if we have a variable without contents.
    """

    token_info = find_token(section, line, col)
    if token_info is not None:
        stack = token_info.stack
        node = token_info.node
        token = token_info.token

        try:
            if (
                token.type == token.ARGUMENT
                and node.__class__.__name__ in CLASSES_WTH_EXPRESSION_ARGUMENTS
            ):
                for part in iter_expression_variables(token):
                    if part.type == token.VARIABLE:
                        if part.col_offset <= col <= part.end_col_offset:
                            return VarTokenInfo(
                                stack, node, part, "$", VarTokenInfo.CONTEXT_EXPRESSION
                            )

                if "$" in token.value:
                    char_in_token = col - token.col_offset
                    if char_in_token >= 0:
                        value_up_to_cursor = token.value[:char_in_token]
                        if value_up_to_cursor.endswith("$"):
                            # Empty variable at this point
                            from robot.api import Token

                            empty_token = Token(
                                type=token.VARIABLE,
                                value="",
                                lineno=token.lineno,
                                col_offset=col,
                            )

                            return VarTokenInfo(
                                stack,
                                node,
                                empty_token,
                                "$",
                                VarTokenInfo.CONTEXT_EXPRESSION,
                            )

        except:
            log.exception("Unable to tokenize: %s", token)

        if "{" in token.value:
            parts = _tokenize_variables_even_when_invalid(token, col)
            if not parts:
                return None

            for part in parts:
                if part.type in (part.VARIABLE, part.ASSIGN):
                    if part.col_offset <= col <= part.end_col_offset:
                        return _find_subvar(
                            token_info.stack, token_info.node, part, col
                        )
            else:
                return None
    return None


def tokenize_variables_from_name(name):
    return tokenize_variables(create_token(name))  # May throw error if it's not OK.


def tokenize_variables(token: IRobotToken) -> Iterator[IRobotToken]:
    return token.tokenize_variables()  # May throw error if it's not OK.


def _tokenize_variables_even_when_invalid(token: IRobotToken, col: int):
    """
    If Token.tokenize_variables() fails, this can still provide the variable under
    the given column by applying some heuristics to find open variables.
    """
    try:
        return token.tokenize_variables()
    except:
        pass

    # If we got here, it means that we weren't able to tokenize the variables
    # properly (probably some variable wasn't closed properly), so, let's do
    # a custom implementation for this use case.

    from robot.api import Token
    from robotframework_ls.impl.robot_constants import VARIABLE_PREFIXES

    diff = col - token.col_offset
    up_to_cursor = token.value[:diff]
    open_at = up_to_cursor.rfind("{")

    if open_at >= 1:
        if up_to_cursor[open_at - 1] in VARIABLE_PREFIXES:
            varname = [up_to_cursor[open_at - 1 :]]
            from_cursor = token.value[diff:]

            for c in from_cursor:
                if c in VARIABLE_PREFIXES or c.isspace() or c == "{":
                    break
                if c == "}":
                    varname.append(c)
                    break
                varname.append(c)

            return [
                Token(
                    type=token.VARIABLE,
                    value="".join(varname),
                    lineno=token.lineno,
                    col_offset=token.col_offset + open_at - 1,
                    error=token.error,
                )
            ]


LIBRARY_IMPORT_CLASSES = ("LibraryImport",)
RESOURCE_IMPORT_CLASSES = ("ResourceImport",)
SETTING_SECTION_CLASSES = ("SettingSection",)


@_convert_ast_to_indexer
def iter_nodes(ast, accept_class: Union[Tuple[str, ...], str]) -> Iterator[NodeInfo]:
    """
    Note: always recursive.
    """
    if not isinstance(accept_class, (list, tuple, set)):
        accept_class = (accept_class,)

    for classname in accept_class:
        yield from ast.iter_indexed(classname)


def iter_all_nodes(ast, recursive=True) -> Iterator[NodeInfo]:
    """
    Note: use this *very* sparingly as no caching will take place
    (as all nodes need to be iterated).

    Use one of the filtered APIs whenever possible as those are cached
    by the type.
    """
    for stack, node in _iter_nodes(ast, recursive=recursive):
        yield NodeInfo(tuple(stack), node)


def is_library_node_info(node_info: NodeInfo) -> bool:
    return node_info.node.__class__.__name__ in LIBRARY_IMPORT_CLASSES


def is_resource_node_info(node_info: NodeInfo) -> bool:
    return node_info.node.__class__.__name__ in RESOURCE_IMPORT_CLASSES


def is_setting_section_node_info(node_info: NodeInfo) -> bool:
    return node_info.node.__class__.__name__ in SETTING_SECTION_CLASSES


@_convert_ast_to_indexer
def iter_library_imports(ast) -> Iterator[NodeInfo[ILibraryImportNode]]:
    yield from ast.iter_indexed("LibraryImport")


@_convert_ast_to_indexer
def iter_resource_imports(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("ResourceImport")


@_convert_ast_to_indexer
def iter_variable_imports(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("VariablesImport")


@_convert_ast_to_indexer
def iter_keywords(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("Keyword")


@_convert_ast_to_indexer
def iter_variables(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("Variable")


@_convert_ast_to_indexer
def iter_tests(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("TestCase")


@_convert_ast_to_indexer
def iter_test_case_sections(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("TestCaseSection")


@_convert_ast_to_indexer
def iter_setting_sections(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("SettingSection")


@_convert_ast_to_indexer
def iter_indexed(ast, clsname) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed(clsname)


def iter_keyword_arguments_as_str(ast, tokenize_keyword_name=False) -> Iterator[str]:
    """
    Provides the arguments with the full representation (use only for getting
    docs).

    May return strings as:
    ${my.store}=${my.load}[first][second]
    """
    for token in _iter_keyword_arguments_tokens(ast, tokenize_keyword_name):
        yield token.value


@_convert_ast_to_indexer
def _iter_keyword_arguments_tokens(
    ast, tokenize_keyword_name=False
) -> Iterator[Tuple[IRobotToken, str]]:
    """
    This API provides tokens as they are.

    i.e.: it returns tokens as: ${my.store}=${my.load}[first][second]

    So, this is internal as the outer world is tailored to
    dealing with what's actually needed.
    """
    for node_info in ast.iter_indexed("Arguments"):
        for token in node_info.node.tokens:
            if token.type == token.ARGUMENT:
                yield token

    if tokenize_keyword_name:
        from robot.api import Token

        ast_node = ast.ast
        if ast_node.__class__.__name__ == "Keyword":
            keyword_name = ast_node.header.get_token(Token.KEYWORD_NAME)
            if keyword_name:
                try:
                    tokenized_vars = keyword_name.tokenize_variables()
                except:
                    pass
                else:
                    for tok in tokenized_vars:
                        if tok.type == Token.VARIABLE:
                            yield tok


def iter_keyword_arguments_as_tokens(
    ast, tokenize_keyword_name=False
) -> Iterator[Tuple[IRobotToken, str]]:
    """
    API tailored at getting argument names.

    It converts an argument such as:
    ${my.store}=${my.load}[first][second]

    and yields something as:

    my.store
    """
    for token in _iter_keyword_arguments_tokens(ast, tokenize_keyword_name):
        for t in _tokenize_subvars(token):
            # The first one is the variable store (the other is the
            # variable load on a default argument)
            # We are only interested in the former in this API.
            yield t
            break  # Just break the inner for.


def iter_keyword_arguments_as_kwarg(
    ast, tokenize_keyword_name=False
) -> Iterator[IKeywordArg]:

    from robotframework_ls.impl.robot_specbuilder import KeywordArg

    for token in _iter_keyword_arguments_tokens(ast, tokenize_keyword_name):
        yield KeywordArg(token.value)


def is_deprecated(ast) -> bool:
    from robotframework_ls.impl.text_utilities import has_deprecated_text

    docs = get_documentation_raw(ast)
    return has_deprecated_text(docs)


def get_documentation_raw(ast: INode) -> str:
    iter_in: Iterator[INode]
    if ast.__class__.__name__ == "File":
        # Handle the case where the File is given (docs must be gotten from
        # the *** Settings *** in this case).
        iter_in = (node_info.node for node_info in iter_setting_sections(ast))
    else:
        iter_in = iter((ast,))

    doc: List[str] = []
    last_line: List[str] = []

    last_token = None
    for ast_node in iter_in:
        for _stack, node in _iter_nodes_filtered_not_recursive(
            ast_node, accept_class="Documentation"
        ):
            for token in node.tokens:
                if last_token is not None and last_token.lineno != token.lineno:
                    doc.extend(last_line)
                    del last_line[:]

                last_token = token

                if token.type in (token.CONTINUATION, token.DOCUMENTATION):
                    # Ignore anything before a continuation.
                    del last_line[:]
                    continue

                last_line.append(token.value)
            else:
                # Last iteration
                doc.extend(last_line)

        if doc:
            # Case with multiple Setting sections
            break

    ret = "".join(doc).strip()
    return ret


def get_documentation_as_markdown(ast) -> str:
    documentation = get_documentation_raw(ast)
    if not documentation:
        return documentation
    try:
        from robotframework_ls import robot_to_markdown

        return robot_to_markdown.convert(documentation)
    except:
        log.exception("Error converting to markdown: %s", documentation)
        return documentation


@_convert_ast_to_indexer
def iter_local_assigns(ast) -> Iterator:
    from robot.api import Token

    for clsname, assign_token_type in (
        ("KeywordCall", Token.ASSIGN),
        ("ForHeader", Token.VARIABLE),
        ("ExceptHeader", Token.VARIABLE),
        ("InlineIfHeader", Token.ASSIGN),
    ):
        for node_info in ast.iter_indexed(clsname):
            node = node_info.node
            for token in node.tokens:
                if token.type == assign_token_type:
                    value = token.value

                    i = value.find("{")
                    j = value.rfind("}")
                    if i != -1 and j != -1 and i >= 1:
                        new_value = value[i + 1 : j]
                        token = Token(
                            type=token.type,
                            value=new_value,
                            lineno=token.lineno,
                            col_offset=token.col_offset + i + 1,
                            error=token.error,
                        )

                        yield VarTokenInfo(node_info.stack, node, token, value[0])


_FIXTURE_CLASS_NAMES = (
    "Setup",
    "Teardown",
    "SuiteSetup",
    "SuiteTeardown",
    "TestSetup",
    "TestTeardown",
)

_CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_TUPLE = _FIXTURE_CLASS_NAMES + (
    "TestTemplate",
    "Template",
)

CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET = frozenset(
    _CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_TUPLE
)

CLASSES_WTH_EXPRESSION_ARGUMENTS = ("IfHeader", "ElseIfHeader", "WhileHeader")


def _tokenize_subvars(initial_token: IRobotToken) -> Iterator[Tuple[IRobotToken, str]]:
    if "{" not in initial_token.value:
        return

    if initial_token.value.startswith("{") and initial_token.value.endswith("}"):
        # i.e.: We're dealing with an expression.
        _first, second, _third = split_token_in_3(
            initial_token,
            "variableOperator",
            "argumentValue",
            "variableOperator",
            1,
            -1,
        )

        for tok in iter_expression_variables(second):
            yield tok, "$"
        return

    skip_type = ""
    robot_match_generator = RobotMatchTokensGenerator(initial_token, skip_type)
    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches

    for robot_match, relative_index in iter_robot_variable_matches(initial_token.value):
        for token in robot_match_generator.gen_tokens_from_robot_match(
            robot_match, relative_index, skip_type, initial_token.type
        ):
            if token.type == initial_token.type:
                found = False
                for v in _tokenize_subvars(token):
                    found = True
                    yield v

                if not found:
                    yield token, robot_match.identifier


@_convert_ast_to_indexer
def iter_variable_references(ast) -> Iterator[VarTokenInfo]:
    # TODO: This right now makes everything globally, we should have 2 versions,
    # one to resolve references which are global and another to resolve references
    # just inside some scope when dealing with local variables.
    # Right now we're not very smart and even if a variable is local we'll reference
    # global variables...

    # Note: we collect only the references, not the definitions here!
    for clsname in (
        "KeywordCall",
        "LibraryImport",
        "ResourceImport",
        "TestTimeout",
        "Variable",
        "ForHeader",
    ) + _FIXTURE_CLASS_NAMES:
        for node_info in ast.iter_indexed(clsname):
            stack = node_info.stack
            node = node_info.node
            token = None
            try:
                for token in node.tokens:
                    if token.type in (token.ARGUMENT, token.NAME):
                        for tok in tokenize_variables(token):
                            if tok.type == token.VARIABLE:

                                # We need to check for inner variables (as in
                                # this case we validate those).
                                for t, var_identifier in _tokenize_subvars(tok):
                                    yield VarTokenInfo(stack, node, t, var_identifier)

            except:
                log.exception("Unable to tokenize: %s", token)

    for clsname in CLASSES_WTH_EXPRESSION_ARGUMENTS:
        for node_info in ast.iter_indexed(clsname):
            stack = node_info.stack
            node = node_info.node
            token = None

            try:
                for token in node.tokens:
                    if token.type == token.ARGUMENT:
                        for tok in iter_expression_variables(token):
                            if tok.type == token.VARIABLE:
                                yield VarTokenInfo(stack, node, tok, "$")
            except:
                log.exception("Unable to tokenize: %s", token)

    for node_info in ast.iter_indexed("Keyword"):
        for token in _iter_keyword_arguments_tokens(
            node_info.node, tokenize_keyword_name=True
        ):
            iter_in = _tokenize_subvars(token)

            try:
                # The first one is the variable store (the other is the
                # variable load on a default argument)
                # We are only interested in the second in this API.
                next(iter_in)
            except StopIteration:
                continue

            for t, varid in iter_in:
                yield VarTokenInfo(stack, node, t, varid)


@_convert_ast_to_indexer
def iter_keyword_usage_tokens(
    ast, collect_args_as_keywords: bool
) -> Iterator[KeywordUsageInfo]:
    """
    Iterates through all the places where a keyword name is being used, providing
    the stack, node, token and name.
    """

    cache_key = ("iter_keyword_usage_tokens", collect_args_as_keywords)
    yield from ast.iter_cached(
        cache_key, _iter_keyword_usage_tokens_uncached, collect_args_as_keywords
    )


def _same_line_col(tok1: IRobotToken, tok2: IRobotToken):
    return tok1.lineno == tok2.lineno and tok1.col_offset == tok2.col_offset


def _build_keyword_usage(
    stack, node, yield_only_for_token, current_tokens
) -> Optional[KeywordUsageInfo]:
    # Note: just check for line/col because the token could be changed
    # (for instance, an EOL ' ' could be added to the token).

    if yield_only_for_token is not None:
        for tok in current_tokens:
            if _same_line_col(yield_only_for_token, tok):
                break
        else:
            return None

    keyword_at_index = 0
    keyword_token = current_tokens[keyword_at_index]
    keyword_token = copy_token_replacing(keyword_token, type=keyword_token.KEYWORD)
    new_tokens = [keyword_token]
    new_tokens.extend(current_tokens[keyword_at_index + 1 :])

    return KeywordUsageInfo(
        stack,
        node.__class__(new_tokens),
        keyword_token,
        keyword_token.value,
        True,
    )


def _iter_keyword_usage_tokens_uncached_from_args(
    stack, node, args_as_keywords_handler, yield_only_for_token=None
):
    # We may have multiple matches, so, we need to setup the appropriate book-keeping
    current_tokens = []

    for token in node.tokens:
        if token.type == token.ARGUMENT:
            next_tok_type = args_as_keywords_handler.next_tok_type(token)

            if next_tok_type in (
                args_as_keywords_handler.CONTROL,
                args_as_keywords_handler.EXPRESSION,
            ):
                # Don't add IF/ELSE IF/AND nor the condition.
                continue

            if next_tok_type != args_as_keywords_handler.KEYWORD:
                # Argument was now added to current_tokens.
                current_tokens.append(token)

                continue

            if current_tokens:
                # Starting a new one (build for the previous).
                usage_info = _build_keyword_usage(
                    stack,
                    node,
                    yield_only_for_token,
                    current_tokens,
                )
                if usage_info is not None:
                    yield usage_info

            current_tokens = [token]

    else:
        # Do one last iteration at the end to deal with the last one.
        if current_tokens:
            usage_info = _build_keyword_usage(
                stack, node, yield_only_for_token, current_tokens
            )
            if usage_info is not None:
                yield usage_info


def _iter_keyword_usage_tokens_uncached(ast, collect_args_as_keywords):
    for clsname in ("KeywordCall",) + _CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_TUPLE:
        for node_info in ast.iter_indexed(clsname):
            stack = node_info.stack
            node = node_info.node
            usage_info = _create_keyword_usage_info(stack, node)
            if usage_info is not None:
                yield usage_info

                if collect_args_as_keywords:
                    args_as_keywords_handler = get_args_as_keywords_handler(
                        usage_info.node
                    )
                    if args_as_keywords_handler is None:
                        continue

                    yield from _iter_keyword_usage_tokens_uncached_from_args(
                        stack, node, args_as_keywords_handler
                    )


def _create_keyword_usage_info(stack, node) -> Optional[KeywordUsageInfo]:
    """
    If this is a keyword usage node, return information on it, otherwise,
    returns None.

    :note: this goes hand-in-hand with get_keyword_name_token.
    """
    from robot.api import Token

    if node.__class__.__name__ == "KeywordCall":
        token_type = Token.KEYWORD

    elif node.__class__.__name__ in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET:
        token_type = Token.NAME

    else:
        return None

    node, token = _strip_node_and_token_bdd_prefix(node, token_type)
    if token is None:
        return None

    keyword_name = token.value
    if keyword_name.lower() == "none":
        return None
    return KeywordUsageInfo(tuple(stack), node, token, keyword_name)


def create_keyword_usage_info_from_token(
    stack: Tuple[INode, ...], node: INode, token: IRobotToken
) -> Optional[KeywordUsageInfo]:
    """
    If this is a keyword usage node, return information on it, otherwise,
    returns None.

    Note that it should return the keyword usage for the whole keyword call
    if we're in an argument that isn't itself a keyword call.
    """
    if token.type == token.ARGUMENT:
        args_as_keywords_handler = get_args_as_keywords_handler(node)
        if args_as_keywords_handler is not None:
            for v in _iter_keyword_usage_tokens_uncached_from_args(
                stack,
                node,
                args_as_keywords_handler,
                yield_only_for_token=token,
            ):
                return v

    return _create_keyword_usage_info(stack, node)


class _ConsiderArgsAsKeywordNames:
    NONE = 0
    KEYWORD = 1
    EXPRESSION = 2
    CONTROL = 3

    def __init__(
        self,
        node,
        normalized_keyword_name,
        consider_keyword_at_index,
        consider_condition_at_index,
    ):
        self._node = node
        self._normalized_keyword_name = normalized_keyword_name
        self._consider_keyword_at_index = consider_keyword_at_index
        self._consider_condition_at_index = consider_condition_at_index
        self._current_arg = 0

        # Run Keyword If is special because it has 'ELSE IF' / 'ELSE'
        # which will then be be (cond, keyword) or just (keyword), so
        # we need to provide keyword usages as needed.
        if self._normalized_keyword_name == "runkeywordif":
            self.next_tok_type = self._next_tok_type_run_keyword_if
        elif self._normalized_keyword_name == "runkeywords":
            found = False
            for token in node.tokens:
                if "AND" == token.value:
                    found = True
                    break
            if found:
                self.next_tok_type = self._next_tok_type_run_keywords
            else:
                self.next_tok_type = self._consider_each_arg_as_keyword

        self._stack_kind = None
        self._stack = None
        self._started_match = False

    def next_tok_type_as_str(self, token) -> str:
        tok_type = self.next_tok_type(token)
        if tok_type == self.NONE:
            return "<none>"
        if tok_type == self.EXPRESSION:
            return "<expression>"
        if tok_type == self.KEYWORD:
            return "<keyword>"
        if tok_type == self.CONTROL:
            return "<control>"
        raise AssertionError(f"Unexpected: {tok_type}")

    def next_tok_type(self, token) -> int:  # pylint: disable=method-hidden
        assert token.type == token.ARGUMENT
        self._current_arg += 1

        if self._current_arg == self._consider_condition_at_index:
            return self.EXPRESSION

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD

        return self.NONE

    def _next_tok_type_run_keyword_if(self, token):
        assert token.type == token.ARGUMENT

        self._current_arg += 1

        if token.value == "ELSE IF":
            self._started_match = True
            self._stack = []
            self._stack_kind = token.value
            return self.CONTROL
        elif token.value == "ELSE":
            self._started_match = True
            self._stack = []
            self._stack_kind = token.value
            return self.CONTROL

        else:
            self._started_match = False
            if self._stack is not None:
                self._stack.append(token)

        if self._stack is not None:
            if self._stack_kind == "ELSE IF":
                if len(self._stack) == 1:
                    return self.EXPRESSION
                return self.KEYWORD if len(self._stack) == 2 else self.NONE

            if self._stack_kind == "ELSE":
                return self.KEYWORD if len(self._stack) == 1 else self.NONE

        if self._current_arg == self._consider_condition_at_index:
            return self.EXPRESSION

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD

        return self.NONE

    def _consider_each_arg_as_keyword(self, token):
        assert token.type == token.ARGUMENT
        return self.KEYWORD

    def _next_tok_type_run_keywords(self, token):
        assert token.type == token.ARGUMENT

        self._current_arg += 1

        if token.value == "AND":
            self._started_match = True
            self._stack = []
            self._stack_kind = token.value
            return self.CONTROL

        else:
            self._started_match = False
            if self._stack is not None:
                self._stack.append(token)

        if self._stack is not None:
            if self._stack_kind == "AND":
                return self.KEYWORD if len(self._stack) == 1 else self.NONE

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD
        return self.NONE


def get_args_as_keywords_handler(node) -> Optional[_ConsiderArgsAsKeywordNames]:
    from robot.api import Token

    if node.__class__.__name__ == "KeywordCall":
        token_type = Token.KEYWORD

    elif node.__class__.__name__ in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET:
        token_type = Token.NAME

    else:
        return None

    node_keyword_name = node.get_token(token_type)
    if node_keyword_name and node_keyword_name.value:
        normalized_keyword_name = normalize_robot_name(node_keyword_name.value)
        consider_keyword_at_index = KEYWORD_NAME_TO_KEYWORD_INDEX.get(
            normalized_keyword_name
        )
        consider_condition_at_index = KEYWORD_NAME_TO_CONDITION_INDEX.get(
            normalized_keyword_name
        )
        if (
            consider_keyword_at_index is not None
            or consider_condition_at_index is not None
        ):
            return _ConsiderArgsAsKeywordNames(
                node,
                normalized_keyword_name,
                consider_keyword_at_index,
                consider_condition_at_index,
            )
    return None


def get_keyword_name_token(
    stack: Tuple[INode, ...], node: INode, token: IRobotToken
) -> Optional[IRobotToken]:
    """
    If the given token is a keyword call name, return the token, otherwise return None.

    :note: this goes hand-in-hand with iter_keyword_usage_tokens.
    """
    if token.type == token.KEYWORD or (
        token.type == token.NAME
        and node.__class__.__name__ in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET
    ):
        return _strip_token_bdd_prefix(token)

    if token.type == token.ARGUMENT and not token.value.strip().endswith("}"):
        args_as_keywords_handler = get_args_as_keywords_handler(node)
        if args_as_keywords_handler is not None:
            for _ in _iter_keyword_usage_tokens_uncached_from_args(
                stack,
                node,
                args_as_keywords_handler,
                yield_only_for_token=token,
            ):
                return token
    return None


def get_library_import_name_token(node, token: IRobotToken) -> Optional[IRobotToken]:
    """
    If the given ast node is a library import and the token is its name, return
    the name token, otherwise, return None.
    """

    if (
        token.type == token.NAME
        and isinstance_name(node, "LibraryImport")
        and node.name == token.value  # I.e.: match the name, not the alias.
    ):
        return token
    return None


def get_resource_import_name_token(node, token: IRobotToken) -> Optional[IRobotToken]:
    """
    If the given ast node is a library import and the token is its name, return
    the name token, otherwise, return None.
    """

    if (
        token.type == token.NAME
        and isinstance_name(node, "ResourceImport")
        and node.name == token.value  # I.e.: match the name, not the alias.
    ):
        return token
    return None


def get_variables_import_name_token(ast, token):
    """
    If the given ast node is a variables import and the token is its name, return
    the name token, otherwise, return None.
    """

    if (
        token.type == token.NAME
        and isinstance_name(ast, "VariablesImport")
        and ast.name == token.value  # I.e.: match the name, not the alias.
    ):
        return token
    return None


def _copy_of_node_replacing_token(node, token, token_type):
    """
    Workaround to create a new version of the same node but with the first
    occurrence of a token of the given type changed to another token.
    """
    new_tokens = list(node.tokens)
    for i, t in enumerate(new_tokens):
        if t.type == token_type:
            new_tokens[i] = token
            break
    return node.__class__(new_tokens)


def _strip_node_and_token_bdd_prefix(node, token_type):
    """
    This is a workaround because the parsing does not separate a BDD prefix from
    the keyword name. If the parsing is improved to do that separation in the future
    we can stop doing this.
    """
    original_token = node.get_token(token_type)
    if original_token is None:
        return node, None
    token = _strip_token_bdd_prefix(original_token)
    if token is original_token:
        # i.e.: No change was done.
        return node, token
    return _copy_of_node_replacing_token(node, token, token_type), token


def _strip_token_bdd_prefix(token):
    """
    This is a workaround because the parsing does not separate a BDD prefix from
    the keyword name. If the parsing is improved to do that separation in the future
    we can stop doing this.

    :return Token:
        Returns a new token with the bdd prefix stripped or the original token passed.
    """
    from robotframework_ls.impl.robot_constants import BDD_PREFIXES
    from robot.api import Token

    if token is None:
        return token

    text = token.value.lower()
    for prefix in BDD_PREFIXES:
        if text.startswith(prefix):
            new_name = token.value[len(prefix) :]
            return Token(
                type=token.type,
                value=new_name,
                lineno=token.lineno,
                col_offset=token.col_offset + len(prefix),
                error=token.error,
            )
    return token


def _append_eol_to_prev_token(last_token, eol_token_contents):
    from robot.api import Token

    new_value = last_token.value + eol_token_contents

    return Token(
        type=last_token.type,
        value=new_value,
        lineno=last_token.lineno,
        col_offset=last_token.col_offset,
        error=last_token.error,
    )


def copy_token_replacing(token, **kwargs):
    from robot.api import Token

    new_kwargs = {
        "type": token.type,
        "value": token.value,
        "lineno": token.lineno,
        "col_offset": token.col_offset,
        "error": token.error,
    }
    new_kwargs.update(kwargs)
    return Token(**new_kwargs)


def copy_token_with_subpart(token, start, end):
    from robot.api import Token

    return Token(
        type=token.type,
        value=token.value[start:end],
        lineno=token.lineno,
        col_offset=token.col_offset + start,
        error=token.error,
    )


def create_range_from_token(token) -> RangeTypedDict:

    start: PositionTypedDict = {"line": token.lineno - 1, "character": token.col_offset}
    end: PositionTypedDict = {
        "line": token.lineno - 1,
        "character": token.end_col_offset,
    }
    code_lens_range: RangeTypedDict = {"start": start, "end": end}
    return code_lens_range


def create_token(name):
    from robot.api import Token

    return Token(Token.NAME, name)


def convert_variable_match_base_to_token(
    token: IRobotToken, variable_match: IRobotVariableMatch
):
    from robot.api import Token

    base = variable_match.base
    assert base is not None
    s = variable_match.string
    if not base:
        base_i = s.find("{") + 1
    else:
        base_i = s.find(base)

    return Token(
        type=token.type,
        value=variable_match.base,
        lineno=token.lineno,
        col_offset=token.col_offset + variable_match.start + base_i,
        error=token.error,
    )


def iter_robot_match_as_tokens(
    robot_match: IRobotVariableMatch, relative_index: int = 0, lineno: int = 0
) -> Iterator[IRobotToken]:
    from robot.api import Token

    base = robot_match.base
    assert base is not None
    s = robot_match.string
    if not base:
        base_i = s.find("{") + 1
    else:
        base_i = s.find(base)

    yield Token(
        type="base",
        value=base,
        lineno=lineno,
        col_offset=relative_index + robot_match.start + base_i,
    )

    last_i = base_i + len(base)
    for item in robot_match.items:
        open_char_i = s.find("[", last_i)
        if open_char_i > 0:
            yield Token(
                type="[",
                value="[",
                lineno=lineno,
                col_offset=relative_index + robot_match.start + open_char_i,
            )

            last_i = open_char_i + 1

        if not item:
            item_i = last_i
        else:
            item_i = s.find(item, last_i)

        yield Token(
            type="item",
            value=item,
            lineno=lineno,
            col_offset=relative_index + robot_match.start + item_i,
        )

        last_i = item_i + len(item)

        close_char_i = s.find("]", last_i)
        if close_char_i < 0:
            break

        yield Token(
            type="]",
            value="]",
            lineno=lineno,
            col_offset=relative_index + robot_match.start + close_char_i,
        )

        last_i = close_char_i


def split_token_in_3(
    token: IRobotToken,
    first_token_type: str,
    second_token_type: str,
    third_token_type,
    start_pos,
    end_pos,
) -> Tuple[IRobotToken, IRobotToken, IRobotToken]:
    first = copy_token_replacing(
        token,
        type=first_token_type,
        value=token.value[:start_pos],
    )
    second = copy_token_replacing(
        token,
        type=second_token_type,
        value=token.value[start_pos:end_pos],
        col_offset=token.col_offset + start_pos,
    )

    if end_pos < 0:
        third = copy_token_replacing(
            token,
            type=third_token_type,
            value=token.value[:end_pos],
            col_offset=second.end_col_offset,
        )
    else:
        third = copy_token_replacing(
            token,
            type=third_token_type,
            value=token.value[end_pos:],
            col_offset=second.end_col_offset,
        )

    return first, second, third


def split_token_change_first(
    token: IRobotToken, first_token_type: str, position: int
) -> Tuple[IRobotToken, IRobotToken]:
    prefix = copy_token_replacing(
        token,
        type=first_token_type,
        value=token.value[:position],
    )
    remainder = copy_token_replacing(
        token, value=token.value[position:], col_offset=prefix.end_col_offset
    )
    return prefix, remainder


def split_token_change_second(
    token: IRobotToken, second_token_type: str, position: int
) -> Tuple[IRobotToken, IRobotToken]:
    prefix = copy_token_replacing(
        token,
        value=token.value[:position],
    )
    remainder = copy_token_replacing(
        token,
        value=token.value[position:],
        col_offset=prefix.end_col_offset,
        type=second_token_type,
    )
    return prefix, remainder


def get_library_arguments_serialized(library) -> Optional[str]:
    return "::".join(library.args) if library.args else None


def iter_expression_variables(expression_token: IRobotToken) -> Iterator[IRobotToken]:
    from robot.api import Token

    for tok in iter_expression_tokens(expression_token):
        if tok.type == "variable":
            yield copy_token_replacing(tok, type=Token.VARIABLE)


class RobotMatchTokensGenerator:
    def __init__(self, token, default_type):
        self.default_type = default_type
        self.token = token
        self.last_gen_end_offset = 0

    def gen_default_type(self, until_offset: int) -> Iterable[IRobotToken]:
        token = self.token
        if until_offset > self.last_gen_end_offset:
            from robot.api import Token

            val = token.value[self.last_gen_end_offset : until_offset]
            if val.strip():  # Don't generate just for whitespaces.
                yield Token(
                    self.default_type,
                    val,
                    token.lineno,
                    token.col_offset + self.last_gen_end_offset,
                    token.error,
                )

    def gen_tokens_from_robot_match(
        self,
        robot_match: IRobotVariableMatch,
        last_relative_index: int,
        op_type: str = "variableOperator",
        var_type: str = "variable",
    ) -> Iterable[IRobotToken]:
        from robot.api import Token

        token = self.token
        if not robot_match.base:
            i = token.value.find("{", robot_match.start + last_relative_index) + 1
        else:
            i = token.value.find(
                robot_match.base, robot_match.start + last_relative_index
            )

        start_offset = robot_match.start + last_relative_index

        yield from self.gen_default_type(start_offset)

        yield Token(
            op_type,
            token.value[robot_match.start + last_relative_index : i],
            token.lineno,
            token.col_offset + start_offset,
            token.error,
        )

        yield Token(
            var_type,
            robot_match.base,
            token.lineno,
            token.col_offset + i,
            token.error,
        )

        base = robot_match.base
        assert base is not None
        j = i + len(base)

        val = token.value[j : robot_match.end + last_relative_index]
        yield Token(
            op_type,
            val,
            token.lineno,
            token.col_offset + j,
            token.error,
        )

        self.last_gen_end_offset = j + len(val)


def iter_expression_tokens(
    expression_token: IRobotToken, default_type="argument"
) -> Iterator[IRobotToken]:
    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches
    from tokenize import generate_tokens, NAME, ERRORTOKEN
    from io import StringIO
    from robot.api import Token

    robot_match_generator = RobotMatchTokensGenerator(expression_token, default_type)

    # Note: tokens yielded are tokens which are internal LS types (not RF types).
    # Use: iter_expression_variables to get the vars with RF types.

    # See: robot.variables.evaluation.evaluate_expression

    iter_robot_match = iter_robot_variable_matches(expression_token.value)
    last_robot_match = None

    try:
        last_robot_match, last_relative_index = next(iter_robot_match)
    except StopIteration:
        pass

    gen_var_token_info: Any = None
    for token_info in generate_tokens(StringIO(expression_token.value).readline):
        if token_info.type == ERRORTOKEN and token_info.string == "$":
            gen_var_token_info = token_info

        elif gen_var_token_info is not None and token_info.type == NAME:
            if gen_var_token_info.start[1] == token_info.start[1] - 1:

                while True:
                    if last_robot_match is None or last_robot_match.base is None:
                        break

                    if (
                        last_robot_match.start + last_relative_index
                        < gen_var_token_info.start[1]
                    ):
                        yield from robot_match_generator.gen_tokens_from_robot_match(
                            last_robot_match, last_relative_index
                        )

                        last_robot_match = None

                        try:
                            last_robot_match, last_relative_index = next(
                                iter_robot_match
                            )
                        except StopIteration:
                            pass
                    else:
                        break

                start_offset = gen_var_token_info.start[1]
                yield from robot_match_generator.gen_default_type(start_offset)

                yield Token(
                    "variableOperator",
                    gen_var_token_info.string,
                    expression_token.lineno,
                    expression_token.col_offset + start_offset,
                    expression_token.error,
                )

                yield Token(
                    "variable",
                    token_info.string,
                    expression_token.lineno,
                    expression_token.col_offset + token_info.start[1],
                    expression_token.error,
                )

                robot_match_generator.last_gen_end_offset = token_info.start[1] + len(
                    token_info.string
                )

            gen_var_token_info = None

    while True:
        if last_robot_match is None or last_robot_match.base is None:
            break

        yield from robot_match_generator.gen_tokens_from_robot_match(
            last_robot_match, last_relative_index
        )

        last_robot_match = None

        try:
            last_robot_match, last_relative_index = next(iter_robot_match)
        except StopIteration:
            pass

    yield from robot_match_generator.gen_default_type(len(expression_token.value))


def is_node_with_expression_argument(node):
    if node.__class__.__name__ == "KeywordCall":
        kw_name = node.keyword
        return kw_name and normalize_robot_name(kw_name) == "evaluate"
    else:
        return node.__class__.__name__ in CLASSES_WTH_EXPRESSION_ARGUMENTS



#!/usr/bin/env python
"""Simple Tk example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for tk
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console
"""

if __name__ == '__main__':
    
    try:
        from Tkinter import *
    except:
        # Python 3
        from tkinter import *
    
    class MyApp:
    
        def __init__(self, root):
            frame = Frame(root)
            frame.pack()
    
            self.button = Button(frame, text="Hello", command=self.hello_world)
            self.button.pack(side=LEFT)
    
        def hello_world(self):
            print("Hello World!")
    
    root = Tk()
    
    app = MyApp(root)



from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IRobotDocument,
    ILibraryDoc,
    IKeywordFound,
    CompletionType,
)
from robocorp_ls_core.lsp import (
    CompletionItemKind,
    TextEditTypedDict,
    CompletionItemTypedDict,
    InsertTextFormat,
)
from typing import Optional, List, Set, Dict, Any
from robotframework_ls.impl.protocols import NodeInfo
import os.path
from robocorp_ls_core import uris
from robocorp_ls_core.protocols import IWorkspace
from robotframework_ls.impl.protocols import ISymbolsCache
from robotframework_ls.impl.robot_constants import ALL_RELATED_FILE_EXTENSIONS


class _Collector(object):
    def __init__(
        self,
        selection,
        token,
        import_location_info: "_ImportLocationInfo",
        imported_keyword_name_to_keyword: Dict[str, List[IKeywordFound]],
    ):
        from robotframework_ls.impl.string_matcher import RobotStringMatcher

        token_str = token.value

        self.completion_items: List[CompletionItemTypedDict] = []
        self.selection = selection
        self.import_location_info = import_location_info
        self.token = token
        self.imported_keyword_name_to_keyword = imported_keyword_name_to_keyword

        self._matcher = RobotStringMatcher(token_str)

    def accepts(self, keyword_name: str) -> bool:
        if not self._matcher.accepts_keyword_name(keyword_name):
            return False

        keywords_found: Optional[
            List[IKeywordFound]
        ] = self.imported_keyword_name_to_keyword.get(keyword_name)
        if not keywords_found:
            return True

        return False

    def _create_completion_item(
        self,
        completion_context: ICompletionContext,
        keyword_name: str,
        selection,
        token,
        col_delta: int,
        memo: Set[str],
        lib_import: Optional[str] = None,
        resource_path: Optional[str] = None,
        data: Optional[Any] = None,
    ) -> Optional[CompletionItemTypedDict]:
        """
        Note: the lib_import and resource_path are the strings to be added
        so that the given library/resource is loaded.

        i.e.: It's the name concatenated to the `Library    {lib_import}` or
        `Resource    {resource_path}`.
        """
        label = f"{keyword_name} ({lib_import or resource_path})"
        if label in memo:
            return None
        memo.add(label)

        prefix = ""
        detail = ""
        import_line = -1
        if completion_context.type != CompletionType.shell:
            if lib_import is not None:
                import_line = self.import_location_info.get_library_import_line()
            elif resource_path is not None:
                import_line = self.import_location_info.get_resource_import_line()

        if import_line == -1:
            # There's no existing import, so, let's see if we have a *** Settings *** section.
            # If we don't we have to create the whole settings, otherwise, we'll add the statement
            # as the first thing in the existing *** Settings *** section.
            if completion_context.type == CompletionType.shell:
                import_line = 0
                prefix = "*** Settings ***\n"
            elif self.import_location_info.setting_section_node_info is None:
                import_line = 0
                prefix = "*** Settings ***\n"
            else:
                import_line = (
                    self.import_location_info.setting_section_node_info.node.end_lineno
                    - 1
                )

        text = keyword_name

        if keyword_name in self.imported_keyword_name_to_keyword:
            check = lib_import or resource_path
            if check:
                basename = os.path.basename(check)
                if basename.endswith(ALL_RELATED_FILE_EXTENSIONS):
                    basename = os.path.splitext(basename)[0]
                text = f"{basename}.{keyword_name}"

        text_edit: TextEditTypedDict = {
            "range": {
                "start": {
                    "line": selection.line,
                    "character": token.col_offset + col_delta,
                },
                "end": {"line": selection.line, "character": token.end_col_offset},
            },
            "newText": text,
        }

        additional_text_edits: List[TextEditTypedDict] = []

        if lib_import is not None:
            additional_text_edits.append(
                {
                    "range": {
                        "start": {"line": import_line, "character": 0},
                        "end": {"line": import_line, "character": 0},
                    },
                    "newText": f"{prefix}Library    {lib_import}\n",
                }
            )
            detail = "* Adds Library Import"
        elif resource_path is not None:
            additional_text_edits.append(
                {
                    "range": {
                        "start": {"line": import_line, "character": 0},
                        "end": {"line": import_line, "character": 0},
                    },
                    "newText": f"{prefix}Resource    {resource_path}\n",
                }
            )
            detail = "* Adds Resource Import"

        completion_item: CompletionItemTypedDict = {
            "label": f"{label}*",
            "detail": detail,
            "kind": CompletionItemKind.Reference,
            "textEdit": text_edit,
            "insertText": text_edit["newText"],
            "insertTextFormat": InsertTextFormat.Snippet,
            "additionalTextEdits": additional_text_edits,
            "data": data,
        }
        self.completion_items.append(completion_item)
        return completion_item


def _collect_auto_import_completions(
    completion_context: ICompletionContext, collector: _Collector
):
    from robotframework_ls.impl.workspace_symbols import iter_symbols_caches
    from robotframework_ls.robot_config import create_convert_keyword_format_func

    symbols_cache: ISymbolsCache
    selection = completion_context.sel
    token = collector.token

    ws: IWorkspace = completion_context.workspace
    folder_paths = []
    for folder in ws.iter_folders():
        folder_paths.append(uris.to_fs_path(folder.uri))

    curr_doc_path = os.path.dirname(uris.to_fs_path(completion_context.doc.uri))

    memo: Set[str] = set()

    default_convert_keyword_format = create_convert_keyword_format_func(
        completion_context.config
    )
    noop = lambda x: x

    for symbols_cache in iter_symbols_caches(
        None, completion_context, show_builtins=False
    ):
        library_info: Optional[ILibraryDoc] = symbols_cache.get_library_info()
        doc: Optional[IRobotDocument] = symbols_cache.get_doc()

        lib_import = None
        resource_path = None

        if library_info is not None:
            if library_info.source:
                if (
                    library_info.source
                    in collector.import_location_info.imported_libraries
                ):
                    continue
            elif library_info.name in collector.import_location_info.imported_libraries:
                continue

            if library_info.source:
                for folder_path in folder_paths:
                    # If the library is found to be in the workspace, use a relative
                    # path, otherwise use the library name (in which case it's expected
                    # to be in the pythonpath).
                    if library_info.source.startswith(folder_path):
                        try:
                            lib_import = os.path.relpath(
                                library_info.source, curr_doc_path
                            ).replace("\\", "/")
                            break
                        except:
                            pass
                else:
                    lib_import = library_info.name

            else:
                lib_import = library_info.name

            convert_keyword_format = default_convert_keyword_format

        elif doc is not None:
            resource_path = doc.path
            try:
                resource_path = os.path.relpath(resource_path, curr_doc_path).replace(
                    "\\", "/"
                )
            except:
                pass
            convert_keyword_format = noop

        for keyword_info in symbols_cache.iter_keyword_info():
            if collector.accepts(keyword_info.name):
                item = collector._create_completion_item(
                    completion_context,
                    convert_keyword_format(keyword_info.name),
                    selection,
                    token,
                    0,
                    memo,
                    lib_import=lib_import,
                    resource_path=resource_path,
                    data=None,
                )
                if item is not None:
                    completion_context.assign_documentation_resolve(
                        item, keyword_info.get_documentation
                    )


class _ImportLocationInfo:
    def __init__(self):
        self.library_node_info: Optional[NodeInfo] = None
        self.resource_node_info: Optional[NodeInfo] = None
        self.setting_section_node_info: Optional[NodeInfo] = None
        self.imported_libraries: Set[str] = set()
        self.imported_resources: Set[str] = set()

    def get_library_import_line(self) -> int:
        if self.library_node_info is not None:
            return self.library_node_info.node.end_lineno
        return -1

    def get_resource_import_line(self) -> int:
        if self.resource_node_info is not None:
            return self.resource_node_info.node.end_lineno
        return -1


def _obtain_import_location_info(completion_context) -> _ImportLocationInfo:
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.libspec_manager import LibspecManager
    from robot.api import Token

    import_location_info = _ImportLocationInfo()

    # Ok, we have something, let's discover where we want to add the
    # 'Library' or 'Resource'.
    ast = completion_context.get_ast()

    libspec_manager: LibspecManager = completion_context.workspace.libspec_manager

    for node_info in ast_utils.iter_nodes(
        ast,
        accept_class=ast_utils.LIBRARY_IMPORT_CLASSES
        + ast_utils.RESOURCE_IMPORT_CLASSES
        + ast_utils.SETTING_SECTION_CLASSES,
    ):
        if ast_utils.is_library_node_info(node_info):
            import_location_info.library_node_info = node_info

            library_name_token = node_info.node.get_token(Token.NAME)
            if library_name_token is not None:
                library_doc_or_error = libspec_manager.get_library_doc_or_error(
                    completion_context.token_value_resolving_variables(
                        library_name_token
                    ),
                    create=True,
                    completion_context=completion_context,
                    args=ast_utils.get_library_arguments_serialized(node_info.node),
                )
                library_doc = library_doc_or_error.library_doc
                if library_doc is not None:
                    if library_doc.source:
                        import_location_info.imported_libraries.add(library_doc.source)
                    else:
                        import_location_info.imported_libraries.add(library_doc.name)

        elif ast_utils.is_resource_node_info(node_info) and node_info.node.name:
            import_location_info.resource_node_info = node_info
            import_location_info.imported_resources.add(node_info.node.name)

        elif ast_utils.is_setting_section_node_info(node_info):
            import_location_info.setting_section_node_info = node_info

    return import_location_info


def complete(
    completion_context: ICompletionContext,
    imported_keyword_name_to_keyword: Dict[str, List[IKeywordFound]],
):
    from robotframework_ls.impl import ast_utils

    token_info = completion_context.get_current_token()
    if token_info is not None:
        token = ast_utils.get_keyword_name_token(
            token_info.stack, token_info.node, token_info.token
        )
        if token is not None:
            import_location_info = _obtain_import_location_info(completion_context)

            collector = _Collector(
                completion_context.sel,
                token,
                import_location_info,
                imported_keyword_name_to_keyword,
            )
            _collect_auto_import_completions(completion_context, collector)

            return collector.completion_items
    return []



import pydevd
import threading

original = pydevd.PyDB.notify_thread_created

found = set()

def new_notify_thread_created(self, thread_id, thread, *args, **kwargs):
    found.add(thread)
    return original(self, thread_id, thread, *args, **kwargs)

pydevd.PyDB.notify_thread_created = new_notify_thread_created 

ok = []
class MyThread(threading.Thread):
    
    def run(self):
        if self not in found:
            ok.append(False)
        else:
            ok.append(True)
        
if __name__ == '__main__':
    threads = []
    for i in range(15):
        t = MyThread()
        t.start()
        threads.append(t)
        
    for t in threads:
        t.join()
        
    assert len(ok) == len(threads)
    assert all(ok), 'Expected all threads to be notified of their creation before starting to run. Found: %s' % (ok,)
    
    found.clear()
    print('TEST SUCEEDED')
        



# Original work Copyright 2008-2015 Nokia Networks
# Original work Copyright 2016-2020 Robot Framework Foundation
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import weakref
from robocorp_ls_core.cache import instance_cache
from typing import Optional, Union, Type, Callable, List, Tuple
from robocorp_ls_core.protocols import Sentinel
from robotframework_ls.impl.protocols import ISymbolsCache, ILibraryDoc, IKeywordArg
from robocorp_ls_core.robotframework_log import get_logger, get_log_level
import json
import typing


log = get_logger(__name__)

_notified_missing_docutils = False


def _rest_to_markdown(doc: str) -> Optional[str]:
    global _notified_missing_docutils
    try:
        import docutils.core
        from robotframework_ls.impl.doctree2md import Writer
    except Exception as e:
        if _notified_missing_docutils:
            return None

        if get_log_level() >= 2:
            log.exception(
                "It's not possible to convert rest to markdown (please make sure that `docutils` is installed)"
            )
        else:
            log.critical(
                "It's not possible to convert rest to markdown (please make sure that `docutils` is installed).\nOriginal error: %s",
                e,
            )

        _notified_missing_docutils = True
        return None

    return docutils.core.publish_string(
        doc,
        writer=Writer(),
        settings_overrides={
            "syntax_highlight": "none",
            "input_encoding": "unicode",
            "output_encoding": "unicode",
        },
    )


def _get_formatter(doc_format) -> Optional[Callable[[str], Optional[str]]]:
    lower_doc_format = doc_format.lower()
    if lower_doc_format == "robot":
        from robotframework_ls import robot_to_markdown

        return robot_to_markdown.convert

    if lower_doc_format == "html":
        from robotframework_ls import html_to_markdown

        return html_to_markdown.convert

    if lower_doc_format == "rest":
        return _rest_to_markdown

    return None


def _markdown_doc(lower_doc_format: str, obj) -> Optional[str]:
    """
    :type obj: LibraryDoc|KeywordDoc

    Note that None is returned if the conversion couldn't be done.
    """
    if not obj.doc:
        return ""

    if lower_doc_format in ("markdown", "md"):
        return obj.doc

    if lower_doc_format == "robot":
        try:
            return obj.__md_doc__
        except AttributeError:
            from robotframework_ls import robot_to_markdown

            obj.__md_doc__ = robot_to_markdown.convert(obj.doc)
        return obj.__md_doc__

    if lower_doc_format == "html":
        try:
            return obj.__md_doc__
        except AttributeError:
            from robotframework_ls import html_to_markdown

            obj.__md_doc__ = html_to_markdown.convert(obj.doc)
        return obj.__md_doc__

    if lower_doc_format == "rest":
        try:
            return obj.__md_doc__
        except AttributeError:
            obj.__md_doc__ = _rest_to_markdown(obj.doc)
        return obj.__md_doc__

    return None


def docs_and_format(obj):
    """
    Given an object with a '.doc_format' and a '.doc', provide a
    tuple with (formatted contents, MarkupKind)

    Where MarkupKind = 'plaintext' | 'markdown'

    Note: may create a `__md_doc__` cache in the object.
    """
    doc_format = obj.doc_format
    lower = doc_format.lower()

    if lower in ("robot", "html", "markdown", "rest"):
        try:
            as_markdown = _markdown_doc(lower, obj)
            if as_markdown is not None:
                return as_markdown, "markdown"

        except:
            log.exception("Error formatting: %s.\nContent:\n%s", lower, obj.doc)

    # Could be 'text' or 'rest' here...
    return obj.doc, "plaintext"


class LibraryDoc(object):
    def __init__(
        self,
        filename,
        name="",
        doc="",
        # This is the RobotFramework version.
        version="",
        # This is the version of the spec.
        specversion="",
        type="library",
        scope="",
        named_args=True,
        doc_format="",
        source=None,
        lineno=-1,
    ):
        assert filename
        self.filename = filename
        self.name = name
        self.doc = doc
        self.version = version
        self.specversion = specversion
        self.type = type
        self.scope = scope
        self.named_args = named_args
        self.doc_format = doc_format or "ROBOT"
        self._source = source
        self.lineno = lineno
        self.inits = []
        self.keywords = []
        self.data_types = []

        self.symbols_cache: Optional[ISymbolsCache] = None

    def to_dictionary(self):
        data_types = {}

        for data_type in self.data_types:
            type_name = data_type.type
            type_name = type_name[0].lower() + type_name[1:] + "s"
            # assert type_name in ("customs", "enums", "typedDicts")
            lst = data_types.setdefault(type_name, [])
            lst.append(data_type.to_dictionary())

        return {
            "name": self.name,
            "doc": self.doc,
            "version": self.version,
            "specversion": self.specversion,
            "type": self.type,
            "scope": self.scope,
            "docFormat": self.doc_format,
            "source": self._source,
            "lineno": self.lineno,
            "tags": list(self.all_tags),
            "inits": [init.to_dictionary() for init in self.inits],
            "keywords": [kw.to_dictionary() for kw in self.keywords],
            "dataTypes": data_types,
        }

    @property  # type: ignore
    @instance_cache
    def source(self):
        # When asked for, make sure that the path is absolute.
        source = self._source
        if source:
            if not os.path.isabs(source):
                source = self._make_absolute(source)
        return source

    @instance_cache
    def _make_absolute(self, source):
        dirname = os.path.dirname(self.filename)
        return os.path.abspath(os.path.join(dirname, source))

    @property
    def doc_format(self):
        return self._doc_format

    @doc_format.setter
    def doc_format(self, doc_format):
        self._doc_format = doc_format or "ROBOT"

    @property
    def keywords(self):
        return self._keywords

    @keywords.setter
    def keywords(self, kws):
        self._keywords = sorted(kws, key=lambda kw: kw.name)

    @property
    def all_tags(self):
        from itertools import chain

        return tuple(chain.from_iterable(kw.tags for kw in self.keywords))

    def __repr__(self):
        return "LibraryDoc(%s, %s, keywords:%s)" % (
            self.filename,
            self.name,
            len(self.keywords),
        )

    __str__ = __repr__

    def convert_docs_to_markdown(self) -> bool:
        old_doc_format = self.doc_format
        formatter = _get_formatter(old_doc_format)

        if formatter is not None:
            new_doc = formatter(self.doc)
            if new_doc is None:
                # We can't format it (docutils not installed?)
                return False

            self.doc = new_doc

            for init in self.inits:
                new = formatter(init.doc)
                if new is not None:
                    init.doc = new

            for keyword in self.keywords:
                new = formatter(keyword.doc)
                if new is not None:
                    keyword.doc = new

            for type_doc in self.data_types:
                new = formatter(type_doc.doc)
                if new is not None:
                    type_doc.doc = new

            self.doc_format = "markdown"

            return True

        return False


class KeywordArg(object):

    POSITIONAL_ONLY = "POSITIONAL_ONLY"
    POSITIONAL_ONLY_MARKER = "POSITIONAL_ONLY_MARKER"
    POSITIONAL_OR_NAMED = "POSITIONAL_OR_NAMED"
    VAR_POSITIONAL = "VAR_POSITIONAL"
    NAMED_ONLY_MARKER = "NAMED_ONLY_MARKER"
    NAMED_ONLY = "NAMED_ONLY"
    VAR_NAMED = "VAR_NAMED"

    _is_keyword_arg = False
    _is_star_arg = False
    _default_value: Union[Type[Sentinel], str] = Sentinel
    _arg_type: Union[Type[Sentinel], str] = Sentinel
    _arg_name: str

    def __init__(
        self,
        arg: str,
        name: Union[Type[Sentinel], str] = Sentinel,
        arg_type: Union[Type[Sentinel], str] = Sentinel,
        default_value: Union[Type[Sentinel], str] = Sentinel,
        kind: str = "",
    ):
        """
        If arg_type and default_value are given, the arg name == arg, otherwise,
        the arg is expected to be something as 'arg:int=10' and thus the arg_type
        and default_value are computed.
        """
        self.original_arg = arg
        self.kind = kind

        if arg.startswith("&"):
            self._is_keyword_arg = True
            if not kind:
                self.kind = self.VAR_NAMED

        elif arg.startswith("@"):
            self._is_star_arg = True
            if not kind:
                self.kind = self.VAR_POSITIONAL

        elif arg.startswith("**"):
            self._is_keyword_arg = True
            arg = "&" + arg[2:]
            if not kind:
                self.kind = self.VAR_NAMED

        elif arg.startswith("*"):
            self._is_star_arg = True
            arg = "@" + arg[1:]
            if not kind:
                self.kind = self.VAR_POSITIONAL

        else:
            if not kind:
                self.kind = self.POSITIONAL_OR_NAMED

            if default_value is not Sentinel:
                self._default_value = default_value
            else:
                eq_i = arg.rfind("=")
                if eq_i != -1:
                    self._default_value = arg[eq_i + 1 :].strip()
                    arg = arg[:eq_i]

            if arg_type is not Sentinel:
                self._arg_type = arg_type
            else:
                if default_value is not Sentinel:
                    # i.e.: if the default value was given this was already
                    # done when we got here.
                    eq_i = arg.rfind("=")
                    if eq_i != -1:
                        arg = arg[:eq_i].strip()

                colon_i = arg.rfind(":")
                if colon_i != -1:
                    self._arg_type = arg[colon_i + 1 :].strip()
                    arg = arg[:colon_i]

        if name is not Sentinel:
            self._arg_name = typing.cast(str, name)
        else:
            self._arg_name = arg

    def to_dictionary(self):
        ret = {
            "name": self._arg_name,
            "kind": self.kind,
            "repr": self.original_arg,
        }
        if self._default_value is not Sentinel:
            ret["defaultValue"] = self._default_value

        if self._arg_type is not Sentinel:
            ret["types"] = self._arg_type

        return ret

    def is_default_value_set(self) -> bool:
        return self._default_value is not Sentinel

    @property
    def arg_name(self) -> str:
        return self._arg_name

    @property
    def is_keyword_arg(self) -> bool:
        return self._is_keyword_arg

    @property
    def is_star_arg(self) -> bool:
        return self._is_star_arg

    def is_arg_type_set(self) -> bool:
        return self._arg_type is not Sentinel

    @property
    def arg_type(self) -> Optional[str]:
        if self._arg_type is Sentinel:
            return None
        return typing.cast(Optional[str], self._arg_type)

    @property
    def default_value(self) -> Optional[str]:
        if self._default_value is Sentinel:
            return None
        return typing.cast(Optional[str], self._default_value)

    def __repr__(self):
        return f"KeywordArg({self.original_arg})"

    __str__ = __repr__


class KeywordDoc(object):
    def __init__(
        self, weak_libdoc, name="", args=(), doc="", tags=(), source=None, lineno=-1
    ):
        self._weak_libdoc = weak_libdoc
        self.name = name
        self._args = args
        self.doc = doc
        self.tags = tags
        self._source = source
        self.lineno = lineno

    @property
    def deprecated(self) -> bool:
        return self.doc.startswith("*DEPRECATED") and "*" in self.doc[1:]

    @property  # type: ignore
    @instance_cache
    def args(self) -> Tuple[IKeywordArg, ...]:
        if self._args:
            if isinstance(self._args[0], KeywordArg):
                return self._args

        return tuple(KeywordArg(arg) for arg in self._args)

    @property  # type: ignore
    @instance_cache
    def source(self) -> str:
        # When asked for, make sure that the path is absolute.
        source = self._source
        if source:
            if not os.path.isabs(source):
                libdoc = self._weak_libdoc()
                if libdoc is not None:
                    source = libdoc._make_absolute(source)
        return source

    @property
    def libdoc(self) -> ILibraryDoc:
        return self._weak_libdoc()

    @property
    def doc_format(self) -> str:
        return self._weak_libdoc().doc_format

    def __repr__(self):
        return "KeywordDoc(%s, line: %s)" % (self.name, self.lineno)

    __str__ = __repr__

    def to_dictionary(self) -> dict:
        return {
            "name": self.name,
            "args": [arg.to_dictionary() for arg in self.args],
            "doc": self.doc,
            "tags": list(self.tags),
            "source": self.source,
            "lineno": self.lineno,
        }

    def __typecheckself__(self) -> None:
        from robotframework_ls.impl.protocols import IKeywordDoc
        from robocorp_ls_core.protocols import check_implements

        _: IKeywordDoc = check_implements(self)


class DataType(object):
    type: str = ""

    def __init__(self, name, doc):
        self.name = name
        self.doc = doc

    def to_dictionary(self):
        return {
            "type": self.type,
            "name": self.name,
            "doc": self.doc,
        }


class TypedDictDoc(DataType):
    type = "TypedDict"

    def __init__(self, name, doc, items=None):
        super().__init__(name, doc)
        self.items = items or []

    def to_dictionary(self):
        return {
            "type": self.type,
            "name": self.name,
            "doc": self.doc,
            "items": self.items,
        }


class EnumDoc(DataType):
    type = "Enum"

    def __init__(self, name, doc, members=None):
        super().__init__(name, doc)
        self.members = members or []

    def to_dictionary(self):
        return {
            "type": self.type,
            "name": self.name,
            "doc": self.doc,
            "members": self.members,
        }


class CustomDoc(DataType):
    type = "Custom"


class SpecDocBuilder(object):
    def build(self, path):
        spec = self._parse_spec(path)

        version = spec.find("version")
        specversion = spec.get("specversion")

        libdoc = LibraryDoc(
            path,
            name=spec.get("name"),
            type=spec.get("type"),
            version=version.text if version is not None else "",
            specversion=specversion if specversion is not None else "",
            doc=spec.find("doc").text or "",
            scope=self._get_scope(spec),
            named_args=self._get_named_args(spec),
            doc_format=spec.get("format", "ROBOT"),
            source=spec.get("source"),
            lineno=int(spec.get("lineno", -1)),
        )

        try:
            specversion = int(specversion)
        except:
            log.exception(f"Error converting specversion: {specversion} to an int.")
            specversion = 0  # Too old?

        if specversion >= 3:
            libdoc.inits = self._create_keywords_v3(
                weakref.ref(libdoc), spec, "inits/init"
            )
            libdoc.keywords = self._create_keywords_v3(
                weakref.ref(libdoc), spec, "keywords/kw"
            )
            try:
                libdoc.data_types = self._create_data_types(spec)
            except:
                log.exception("Error loading data types from libspec.")
        else:
            libdoc.inits = self._create_keywords_v2(weakref.ref(libdoc), spec, "init")
            libdoc.keywords = self._create_keywords_v2(weakref.ref(libdoc), spec, "kw")
        return libdoc

    def _get_scope(self, spec):
        # RF >= 3.2 has "scope" attribute w/ value 'GLOBAL', 'SUITE, or 'TEST'.
        if "scope" in spec.attrib:
            return spec.get("scope")
        # RF < 3.2 has "scope" element. Need to map old values to new.
        scope = spec.find("scope").text
        return {
            "": "GLOBAL",  # Was used with resource files.
            "global": "GLOBAL",
            "test suite": "SUITE",
            "test case": "TEST",
        }[scope]

    def _create_data_types(self, spec):
        data_types = [
            self._create_enum_doc(dt) for dt in spec.findall("datatypes/enums/enum")
        ]
        data_types.extend(
            self._create_typed_dict_doc(dt)
            for dt in spec.findall("datatypes/typeddicts/typeddict")
        )
        data_types.extend(
            self._create_custom_doc(dt)
            for dt in spec.findall("datatypes/customs/custom")
        )
        return data_types

    def _create_enum_doc(self, dt):
        return EnumDoc(
            name=dt.get("name"),
            doc=dt.find("doc").text or "",
            members=[
                {"name": member.get("name"), "value": member.get("value")}
                for member in dt.findall("members/member")
            ],
        )

    def _create_typed_dict_doc(self, dt):
        items = []
        for item in dt.findall("items/item"):
            required = item.get("required", None)
            if required is not None:
                required = True if required == "true" else False
            items.append(
                {"key": item.get("key"), "type": item.get("type"), "required": required}
            )
        return TypedDictDoc(
            name=dt.get("name"), doc=dt.find("doc").text or "", items=items
        )

    def _create_custom_doc(self, dt):
        return CustomDoc(name=dt.get("name"), doc=dt.find("doc").text or "")

    def _parse_spec(self, path):
        try:
            from xml.etree import cElementTree as ET
        except ImportError:
            from xml.etree import ElementTree as ET

        if not os.path.isfile(path):
            raise IOError("Spec file '%s' does not exist." % path)
        root = ET.parse(path).getroot()
        if root.tag != "keywordspec":
            raise RuntimeError("Invalid spec file '%s'." % path)
        return root

    def _get_named_args(self, spec):
        elem = spec.find("namedargs")
        if elem is None:
            return False  # Backwards compatiblity with RF < 2.6.2
        return elem.text == "yes"

    # ===========================================================================
    # V2 handling
    # ===========================================================================
    def _create_keywords_v2(self, weak_libdoc, spec, path):
        ret = []
        for elem in spec.findall(path):
            args = []
            for a in elem.findall("arguments/arg"):
                if a.text == "*":
                    continue
                args.append(a.text)
            ret.append(
                KeywordDoc(
                    weak_libdoc,
                    name=elem.get("name", ""),
                    args=tuple(args),
                    doc=elem.find("doc").text or "",
                    tags=tuple(t.text for t in elem.findall("tags/tag")),
                    source=elem.get("source"),
                    lineno=int(elem.get("lineno", -1)),
                )
            )
        return ret

    # ===========================================================================
    # V3 handling
    # ===========================================================================
    def _create_arguments_v3(self, elem):
        ret = []
        for arg in elem.findall("arguments/arg"):
            name = arg.find("name")
            if name is None:
                continue
            name = name.text

            arg_repr = arg.get("repr")
            if not arg_repr:
                arg_repr = name

            kind = arg.get("kind")
            if not kind or kind in ("VAR_POSITIONAL", "VAR_NAMED"):
                # Default handling for *args and **kwargs converts to &args / @args
                ret.append(KeywordArg(arg_repr, kind=kind))
                continue

            arg_type = arg.find("type")
            arg_default = arg.find("default")
            ret.append(
                KeywordArg(
                    arg_repr,
                    name,
                    arg_type.text if arg_type is not None else Sentinel,
                    arg_default.text if arg_default is not None else Sentinel,
                    kind=kind,
                )
            )

        return ret

    def _create_keywords_v3(self, weak_libdoc, spec, path):
        ret = []
        for elem in spec.findall(path):
            ret.append(
                KeywordDoc(
                    weak_libdoc,
                    name=elem.get("name", ""),
                    args=tuple(self._create_arguments_v3(elem)),
                    doc=elem.find("doc").text or "",
                    tags=[t.text for t in elem.findall("tags/tag")],
                    source=elem.get("source"),
                    lineno=int(elem.get("lineno", -1)),
                )
            )
        return ret


class JsonDocBuilder:
    def build(self, path):
        assert isinstance(path, str)
        spec = self._parse_spec_json(path)
        return self.build_from_dict(path, spec)

    def build_from_stream(self, spec_filename, stream):
        spec = json.loads(stream.read())
        return self.build_from_dict(spec_filename, spec)

    def build_from_dict(self, filename, spec):
        libdoc = LibraryDoc(
            filename,
            name=spec["name"],
            doc=spec["doc"],
            version=spec["version"],
            specversion=spec["specversion"],
            type=spec["type"],
            scope=spec["scope"],
            doc_format=spec["docFormat"],
            source=spec["source"],
            lineno=int(spec.get("lineno", -1)),
        )
        new_data_types = []

        for custom in spec["dataTypes"].get("customs", []):
            new_data_types.append(
                CustomDoc(
                    name=custom["name"],
                    doc=custom["doc"],
                )
            )

        for enum in spec["dataTypes"].get("enums", []):
            new_data_types.append(
                EnumDoc(
                    name=enum["name"],
                    doc=enum["doc"],
                    members=enum["members"],
                )
            )
        for typed_dict in spec["dataTypes"].get("typedDicts", []):
            new_data_types.append(
                TypedDictDoc(
                    name=typed_dict["name"],
                    doc=typed_dict["doc"],
                    items=typed_dict["items"],
                )
            )

        libdoc.data_types = new_data_types

        weak_libdoc = weakref.ref(libdoc)
        libdoc.inits = [self._create_keyword(kw, weak_libdoc) for kw in spec["inits"]]
        libdoc.keywords = [
            self._create_keyword(kw, weak_libdoc) for kw in spec["keywords"]
        ]
        return libdoc

    def _parse_spec_json(self, path):
        if not os.path.isfile(path):
            raise RuntimeError("Spec file '%s' does not exist." % path)
        with open(path) as json_source:
            libdoc_dict = json.load(json_source)
        return libdoc_dict

    def _create_keyword(self, kw, weak_libdoc):
        return KeywordDoc(
            name=kw.get("name"),
            args=self._create_arguments(kw["args"]),
            doc=kw["doc"],
            tags=kw["tags"],
            source=kw["source"],
            lineno=int(kw.get("lineno", -1)),
            weak_libdoc=weak_libdoc,
        )

    def _create_arguments(self, arguments) -> List[KeywordArg]:
        new_arguments = []

        for argument in arguments:
            arg = argument["repr"]
            name = argument["name"]
            kind = argument["kind"]

            kwargs = {
                "arg": arg,
                "name": name,
                "kind": kind,
            }

            arg_type = argument.get("types")
            if arg_type is not None:
                kwargs["arg_type"] = arg_type

            default_value = argument.get("defaultValue")
            if default_value is not None:
                kwargs["default_value"] = default_value
            new_arguments.append(KeywordArg(**kwargs))

        return new_arguments



import sys
import os

try:
    from _debugger_case_dont_trace import call_me_back
except ImportError:
    sys.path.append(os.path.dirname(__file__))
    from _debugger_case_dont_trace import call_me_back


def my_callback():
    print('trace me')  # Break here


if __name__ == '__main__':
    call_me_back(my_callback)
    call_me_back(my_callback)
    print('TEST SUCEEDED!')



import warnings


class MyClass(object):

    def __getattribute__(self, attr):
        warnings.warn(
            "Deprecation Warning",
            DeprecationWarning
        )

        warnings.warn(
            "Future Warning!",
            FutureWarning,
        )
        return 1


obj = MyClass()

if __name__ == '__main__':
    print('TEST SUCEEDED')  # break here



import time
import threading


class ProceedContainer:
    tid_to_proceed = {
        1: False,
        2: False,
    }


def exit_while_loop(tid):
    ProceedContainer.tid_to_proceed[tid] = True
    return 'ok'


def thread_func(tid):
    while not ProceedContainer.tid_to_proceed[tid]:  # The debugger should change the proceed to True to exit the loop.
        time.sleep(.1)


if __name__ == '__main__':
    threads = [
        threading.Thread(target=thread_func, args=(1,)),
        threading.Thread(target=thread_func, args=(2,)),
    ]
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    print('TEST SUCEEDED')
